from .helper import *
from .metrics import *
from .image import Image, _test_stitched_data
from .config import *
import timeit
import glob, os, tifffile
import pickle as pk
from scipy import stats
import multiprocessing as mp
from matplotlib import rc
from matplotlib.gridspec import GridSpec


def plot_cell(cell, channel=None, savefig=False,
              filename=None, scale_bar=False,
              lw=0.5, return_fig=False,
              c_midline='royalblue',
              c_contour='orange',
              d_range=[],
              preserve_scale=False,
              scale_bar_weight=3,
              dpi=160):

    height, width = cell.mask.shape
    if preserve_scale or savefig:
        fig, ax = get_img_figure(cell.data[cell.mask_channel_name], dpi)
    else:
        fig, ax = plt.subplots(figsize=(4, 4))

    # step1, parameter initialization
    one_micron_l = 1 / cell.pixel_microns

    # step2 determine cell orientation
    # find x,y coordinates of cell mask
    xcoor, ycoor = np.where(cell.mask > 0)
    # locate far-left, far-right indices
    ymin_id = np.argmin(ycoor)
    ymax_id = np.argmax(ycoor)
    x_left, x_right = xcoor[[ymin_id, ymax_id]]
    if x_left <= x_right:
        # left side tilting upwards
        scalebar_y = 0.1 * height
    else:
        # right side tilting upwards
        scalebar_y = 0.8 * height

    # draw scale bar
    start = width * 0.9 - one_micron_l
    end = width * 0.9
    if scale_bar:
        ax.hlines(y=scalebar_y, xmin=start, xmax=end, color='white', lw=scale_bar_weight)

    # plot data
    if channel == None:
        ax.imshow(cell.data[cell.mask_channel_name], cmap='gist_gray')
    elif channel=='Shape_index':
        ax.imshow(cell.shape_index, cmap='viridis')
    else:
        if len(d_range) == 0:
            ax.imshow(cell.data[channel], cmap='gist_gray')
        else:
            ax.imshow(cell.data[channel], cmap='gist_gray',
                       vmin=d_range[0], vmax=d_range[1])

    # plot contour for branched object
    if cell.discarded or cell.branched:
        plt.plot(cell.init_contour.T[1],
                 cell.init_contour.T[0], c='salmon', lw=lw)

    # plot contour for regular object
    else:
        plt.plot(cell.optimized_contour.T[1],
                 cell.optimized_contour.T[0], c=c_contour, lw=lw)

    # plot midline
    for midline in cell.midlines:
        plt.plot(midline.T[1], midline.T[0], c=c_midline, lw=lw)

    plt.axis('off')

    if savefig:
        fig.savefig(filename)
        plt.close(fig)
    elif return_fig:
        return ax


def plot_puncta(cell, channel, savefig=False, filename=None, lw=1, dpi=160):

    fig, ax = get_img_figure(cell.data[cell.mask_channel_name], dpi)

    if channel == cell.mask_channel_name:
        raise ValueError('Puncta function is not supported for non-fluorescent images!')
    else:
        ax.imshow(cell.data[channel], cmap='gist_gray')

    if cell.discarded or cell.branched:
        raise ValueError('Abnormal cell!')
    else:
        ax.plot(cell.optimized_contour.T[1], cell.optimized_contour.T[0], c='orange', lw=lw)
        puncta = cell.fluorescent_puncta[channel]
        if len(puncta) > 0:
            xcoor = puncta[:, 0]
            ycoor = puncta[:, 1]
            intensity = np.log2(puncta[:, 2])
            ax.scatter(ycoor, xcoor, s=intensity * 30, ec='royalblue', lw=2, fc='None')
            ax.scatter(ycoor, xcoor, s=5, color='red')
    plt.axis('off')
    if savefig:
        fig.savefig(filename)
        plt.close(fig)


def cell_plot_full(cell,
                   savefig=False,
                   filename=None,
                   dpi=240):
    n_channels = len(cell.data)
    r = 4 * (1 + 2 * (n_channels - 1))
    c = 20
    fig = plt.figure(figsize=(c / 2, r / 2))
    gs = GridSpec(r, c)

    ref_channel = cell.mask_channel_name
    data = cell.measurements.signal_measurements
    width = cell.width_lists[0]
    length = cell.lengths[0]

    # plot ref data
    ax0 = fig.add_subplot(gs[:4, :16])
    ref_data = -data[ref_channel]['axial']
    ref_data = (ref_data - ref_data.min()) / (ref_data.max() - ref_data.min())
    ref_corr = np.linspace(0, length, len(ref_data))
    ax0.plot(ref_corr, ref_data, lw=3, ls='--', color='salmon', label='reversed phase contrast')
    ax0.set_ylim(0, 1.2)
    ax0.set_xlim(0,length)
    ax0.set_ylabel('normalized,\nreversed\nphase-contrast', fontname='Arial', fontsize=12)
    ax0.set_yticks([0.2, 0.6, 1.0])
    ax0.set_yticklabels([0.2, 0.6, 1.0], fontname='Arial', fontsize=12)
    ax0.get_xaxis().set_visible(False)
    plt.legend(loc=(0.12, 0.1), prop=rc('font', family='Arial'), fontsize=12)

    ax1 = ax0.twinx()
    ax1.plot(ref_corr, width * cell.pixel_microns, lw=3,
             color=(33 / 255, 179 / 255, 214 / 255, 1), label='cell width')
    ax1.set_ylim(0, 1.5)
    ax1.set_ylabel('cell width [µm]', fontname='Arial', fontsize=12, rotation=270, va='bottom')
    ax1.set_yticks([0.1, 0.75, 1.4])
    ax1.set_yticklabels([0.1, 0.5, 0.9], fontname='Arial', fontsize=12)
    ax1.get_xaxis().set_visible(False)
    plt.legend(loc=(0.6, 0.1), prop=rc('font', family='Arial'), fontsize=12)
    # ax2 = fig.add_subplot(gs[:4,16:])
    # ax2.get_xaxis().set_visible(False)
    # ax2.get_yaxis().set_visible(False)
    counter = 0

    for channel, img in cell.data.items():
        if channel != ref_channel:
            ax_plot = fig.add_subplot(gs[4 + counter * 4:8 + counter * 4, :16])
            ax_img = fig.add_subplot(gs[8 + counter * 4:12 + counter * 4, :16])
            ax_lateral = fig.add_subplot(gs[8 + counter * 4:12 + counter * 4, 16:])

            if counter == 0:
                ax_text = fig.add_subplot(gs[4 + counter * 4:8 + counter * 4, 16:])
                ax_text.text(x=0.05, y=0.8, s='Cell_id: {}-{}'.format(cell.cluster_label, cell.cell_label),
                             fontname='Arial', fontsize=12)
                ax_text.text(x=0.05, y=0.6, s='length [µm]: {}'.format(round(length, 2)),
                             fontname='Arial', fontsize=12)
                ax_text.text(x=0.05, y=0.4, s='width [µm]: {}'.format(round(np.median(width * cell.pixel_microns), 2)),
                             fontname='Arial', fontsize=12)

            counter += 2

            ax_plot.plot(np.linspace(0, 1, len(data[channel]['axial_mean'])),
                         data[channel]['axial_mean'], lw=3, color=(33 / 255, 179 / 255, 214 / 255, 1))
            ax_plot.fill_between(x=np.linspace(0, 1, len(data[channel]['axial_mean'])),
                                 y1=data[channel]['axial_mean'] - data[channel]['axial_std'],
                                 y2=data[channel]['axial_mean'] + data[channel]['axial_std'],
                                 color=(33 / 255, 179 / 255, 214 / 255, 1), alpha=0.2)
            ax_plot.set_xlim(0, 1)
            ax_plot.set_ylabel('{} intensity'.format(channel), fontname='Arial', fontsize=12)
            ax_plot.set_yticklabels(ax_plot.get_yticks(), fontname='Arial', fontsize=12)
            ax_plot.get_xaxis().set_visible(False)

            ax_img.imshow(data[channel]['straighten_normalized'], aspect='auto')
            ax_img.get_xticks()
            xticks = (ax_img.get_xticks())
            newxticks = [np.round(length * (tick / data[channel]['straighten_normalized'].shape[1]), 1) for tick in
                         xticks]
            ax_img.set_xticklabels(newxticks, fontname='Arial', fontsize=12)
            ax_img.set_yticks([])
            ax_img.set_ylabel('{}\nstraighten image\n'.format(channel), fontname='Arial', fontsize=12)
            ax_img.set_xlabel('cell length [µm]'.format(channel), fontname='Arial', fontsize=12)

            ax_text.get_xaxis().set_visible(False)
            ax_text.get_yaxis().set_visible(False)

            lateral_mean = data[channel]['lateral_mean']
            lateral_std = data[channel]['lateral_std']

            ax_lateral.plot(lateral_mean, np.linspace(0, 1, len(lateral_mean)), lw=3,
                            color=(33 / 255, 179 / 255, 214 / 255, 1))
            ax_lateral.fill_betweenx(y=np.linspace(0, 1, len(lateral_mean)), color=(33 / 255, 179 / 255, 214 / 255, 1),
                                     x1=lateral_mean - lateral_std, x2=lateral_mean + lateral_std, alpha=0.2)
            ax_lateral.get_yaxis().set_visible(False)

            lateral_ticks = ax_lateral.get_xticks()
            lateral_ticks = [lateral_ticks[0], lateral_ticks[-1]]
            ax_lateral.set_xticks(lateral_ticks)
            ax_lateral.set_xticklabels(lateral_ticks, fontname='Arial', fontsize=12)
            ax_lateral.set_xlabel('{} lateral axis'.format(channel), fontname='Arial', fontsize=12)

    if savefig:
        fig.savefig(filename, bbox_inches='tight', dpi=dpi, transparent=True)
        plt.close(fig)


def plot_segmented_cluster(image,
                           cluster_key,
                           channel='default',
                           savefig=False,
                           filename=None,
                           show_branched=True,
                           show_accepted=True,
                           show_discarded=True,
                           c_discarded='black',
                           c_branched='crimson',
                           c_accepted='orange',
                           c_midline='royalblue',
                           line_style='-',
                           lw=1,
                           scale_bar=False,
                           scale_bar_length=10,
                           scale_bar_weight=6,
                           dpi=240,
                           preserve_scale=False):

    cluster = image.clusters[cluster_key]
    if preserve_scale or savefig:
        height, width = cluster.mask.shape
        fig, ax = get_img_figure(cluster.mask, dpi)
    else:
        fig, ax = plt.subplots(figsize=(5, 5))
    xc1, yc1, xc2, yc2 = cluster.bbox

    if channel == 'default':
        canvas = image.data[image.mask_channel_name][xc1:xc2, yc1:yc2]
    elif channel in image.data:
        canvas = image.data[channel][xc1:xc2, yc1:yc2]
    else:
        raise ValueError('No {} channel found!'.format(channel))

    plt.imshow(canvas, cmap='gist_gray')
    for cell_id, cell in cluster.cells.items():
        x1, y1, x2, y2 = cell.bbox
        x1 -= xc1
        x2 -= xc1

        y1 -= yc1
        y2 -= yc1
        if cell.discarded:
            if show_discarded:
                contour = cell.init_contour
                plt.plot(contour[:, 1] + y1, contour[:, 0] + x1, lw=lw, c=c_discarded, ls=line_style)
        elif cell.branched:
            if show_branched:
                contour = cell.init_contour
                plt.plot(contour[:, 1] + y1, contour[:, 0] + x1, lw=lw, c=c_branched, ls=line_style)
        else:
            if show_accepted:
                contour = cell.optimized_contour
                midline = cell.midlines[0]
                plt.plot(contour[:, 1] + y1, contour[:, 0] + x1, lw=lw, c=c_accepted, ls=line_style)
                plt.plot(midline[:, 1] + y1, midline[:, 0] + x1, lw=lw, c=c_midline, ls=line_style)
    plt.axis('off')

    # plot scale bar:
    if scale_bar:
        height, width = cluster.mask.shape
        micron_l = scale_bar_length / image.pixel_microns
        scalebar_y = 0.9 * height
        start = width * 0.9 - micron_l
        end = width * 0.9
        if scale_bar:
            plt.hlines(y=scalebar_y, xmin=start, xmax=end, color='white', lw=scale_bar_weight)

    if savefig:
        plt.savefig(filename)
        plt.close(fig)


def plot_segmented_image(image,
                         channel='default',
                         savefig=False,
                         filename=None,
                         show_branched=False,
                         show_accepted=True,
                         show_discarded=False,
                         c_discarded='black',
                         c_branched='crimson',
                         c_accepted='orange',
                         c_midline='royalblue',
                         lw=0.5,
                         dpi=240,
                         scale=1):

    fig, ax = get_img_figure(image.data[image.mask_channel_name], dpi, scale=scale)
    if channel == 'default':
        canvas = image.data[image.mask_channel_name]
    elif channel in image.data:
        canvas = image.data[channel]
    else:
        raise ValueError('No {} channel found!'.format(channel))
    ax.imshow(canvas, cmap='gist_gray')
    for cluster_id, cluster in image.clusters.items():
        if cluster.cells != {}:
            for cell_id, cell in cluster.cells.items():
                if cell.discarded:
                    if show_discarded:
                        contour = cell.init_contour
                        x1, y1, x2, y2 = cell.bbox
                        ax.plot(contour[:, 1]+y1, contour[:, 0]+x1, lw=lw, c=c_discarded)
                elif cell.branched:
                    if show_branched:
                        contour = cell.init_contour
                        x1, y1, x2, y2 = cell.bbox
                        ax.plot(contour[:, 1] + y1, contour[:, 0] + x1, lw=lw, c=c_branched)
                else:
                    if show_accepted:
                        contour = cell.optimized_contour
                        midline = cell.midlines[0]
                        x1, y1, x2, y2 = cell.bbox
                        ax.plot(contour[:, 1] + y1, contour[:, 0] + x1, lw=lw, c=c_accepted)
                        ax.plot(midline[:, 1] + y1, midline[:, 0] + x1, lw=lw, c=c_midline)
    plt.axis('off')
    if savefig:
        fig.savefig(filename)
        plt.close(fig)


def plot_segmented_binary(image, savefig=False, filename=None):

    canvas = np.zeros(image.mask_binary.shape)
    for cluster_id, cluster in image.clusters.items():
        if cluster.cells != {}:
            for cell_id, cell in cluster.cells.items():
                if not cell.discarded and not cell.branched:
                    x1, y1, x2, y2 = cell.bbox
                    canvas[x1:x2, y1:y2][cell.mask>0] = 255
    canvas = canvas.astype(np.uint8)
    if savefig:
        tifffile.imsave(filename, canvas, imagej=True)
    else:
        tifffile.imshow(canvas)


def get_img_figure(image, dpi, scale=1):
    """
    Create a matplotlib (figure,axes) for an image (numpy array) setup so that
        a) axes will span the entire figure (when saved no whitespace)
        b) when saved the figure will have the same x/y resolution as the array,
           with the dpi value you pass in.

    Arguments:
        image -- numpy 2d array
        dpi -- dpi value that the figure should use

    Returns: (figure, ax) tuple from plt.subplots
    """

    # get required figure size in inches (reversed row/column order)
    inches = (scale*image.shape[1]/dpi, scale*image.shape[0]/dpi)

    # make figure with that size and a single axes
    fig, ax = plt.subplots(figsize=inches, dpi=dpi)

    # move axes to span entire figure area
    fig.subplots_adjust(left=0, right=1, top=1, bottom=0, wspace=0, hspace=0)

    return fig, ax


def demograph(data_dict_file, channel,
              max_length=8, pixel_width=512,
              filename = None, width=3, height=4,
              save_img=False, dpi=240):
    data_dict = pk.load(open(data_dict_file,'rb'))
    cell_id = data_dict['sorted_filtered_id']
    data = data_dict[channel]['padded_axial_data'][cell_id]

    fig = plt.figure(figsize=(width,height))
    ax = plt.subplot(111)
    ax.imshow(data,cmap='viridis',aspect='auto')
    pos = [int(pixel_width*0.25)*x for x in range(5)]
    label = [int(max_length*0.25)*x for x in range(5)]
    plt.xticks(pos,label)
    plt.xlabel('length [um]')
    ax.yaxis.set_visible(False)
    if save_img:
        plt.savefig(filename, dpi=dpi, bbox_inches='tight')
        plt.close


def create_canvas(width=101, height=600):
    canvas = np.zeros((width, height))
    r = int(width / 2)
    rr1, cc1 = draw.ellipse(r, int(r * 1.6), r, int((r) * 1.6))
    rr2, cc2 = draw.ellipse(r, int(height - 1.6 * r), r, int((r) * 1.6))
    rr3, cc3 = draw.rectangle(start=(1, int(r * 1.6)), end=(width - 2, height - int(r * 1.6)),
                                        shape=canvas.shape)
    canvas[rr3, cc3] = 1
    canvas[rr1, cc1] = 1
    canvas[rr2, cc2] = 1
    l = len(np.nonzero(np.sum(canvas, axis=0))[0])
    counter = 0
    canvas = canvas.T
    return canvas


def create_mesh(canvas):
    xt, yt = np.nonzero(canvas)
    l, m = np.sum(canvas, axis=0), np.sum(canvas, axis=1)
    norm_yt = np.zeros(xt.shape)
    norm_xt = (xt - xt.min()) / (l.max() - 1)
    count = 0
    for i in range(len(xt)):
        r = m.max() / m[xt[i]]
        norm_yt[i] = count * r / (m.max() - 1)
        if i != len(xt) - 1:
            if xt[i + 1] == xt[i]:
                count += 1
            else:
                count = 0
        else:
            count += 1
    return (xt, yt, norm_xt, norm_yt)


def project_image(xt, yt, norm_xt, norm_yt, canvas, data):
    paint = np.zeros(canvas.shape)
    xid = norm_xt.copy()
    yid = norm_yt.copy()
    xid *= data.shape[0] - 1
    yid *= data.shape[1] - 1
    interpolated = bilinear_interpolate_numpy(data, xid, yid)
    paint[xt, yt] = interpolated
    return (paint)


def group_by_length(length, groups=[3, 4, 5, 6]):
    if length < groups[0]:
        return 1
    elif length < groups[1]:
        return 2
    elif length < groups[2]:
        return 3
    elif length < groups[3]:
        return 4
    else:
        return 5


def initiate_projection():
    width = 75
    heights = [250, 350, 450, 550, 650]
    pad = np.zeros((50, width))
    gap = np.zeros((750, 10))
    padded = [gap]
    xt_list, yt_list, nxt_list, nyt_list = [], [], [], []
    for i in range(len(heights)):
        a = create_canvas(width=width, height=heights[i])
        half_pad = np.tile(pad, (len(heights) - i, 1))
        m_pad = np.concatenate([half_pad, a, half_pad], axis=0)
        m_pad = np.concatenate([m_pad, gap], axis=1)
        xt, yt, norm_xt, norm_yt = create_mesh(m_pad)
        xt_list.append(xt)
        yt_list.append(yt)
        nxt_list.append(norm_xt)
        nyt_list.append(norm_yt)
        padded.append(m_pad)
    contours = measure.find_contours(np.concatenate(padded, axis=1), level=0)
    optimized_outline = []
    for contour in contours:
        optimized_outline.append(spline_approximation(contour, n=2 * len(contour)))
    return padded, xt_list, yt_list, nxt_list, nyt_list, optimized_outline


def divide_by_fraction(data_length, by=5):
    out = []
    for i in range(by - 1):
        out.append(i * int(data_length / by))
    out.append(data_length)
    return out


def project_by_percentile(length_sorted_index, cells, channel,
                          canvas=None, re_align=True, min_cells=10,
                          savefig=False, filename=None, hide_title=True):

    #
    if canvas == None:
        padded, xt_list, yt_list, nxt_list, nyt_list, optimized_outline = initiate_projection()
    else:
        padded, xt_list, yt_list, nxt_list, nyt_list, optimized_outline = canvas

    well_paints = padded.copy()
    for m in well_paints:
        m *= 0
    group_count = [0, 0, 0, 0, 0]

    grouped_index_fraction = divide_by_fraction(len(length_sorted_index), by=6)
    for i in range(5):
        p1 = grouped_index_fraction[i]
        p2 = grouped_index_fraction[i + 1]
        index_list = length_sorted_index[p1:p2]

        for idx in index_list:
            cell = cells[idx]

            group_count[i] += 1
            group = i + 1
            canvas = padded[group].copy()

            img = cell.data[channel]
            midline = cell.midlines[0]
            width = cell.width_lists[0]
            data = straighten_cell_normalize_width(img, midline, width)

            normalized_data = normalize_data_2D(data, re_align=re_align)

            well_paints[group] += project_image(xt_list[group - 1], \
                                                yt_list[group - 1], \
                                                nxt_list[group - 1], \
                                                nyt_list[group - 1], \
                                                canvas, normalized_data.T)

    for j in range(len(well_paints)):
        graph = well_paints[j]
        if graph.max() > 0:
            min_val = graph[graph > 0].min()
            normalized = (graph - min_val) / (graph.max() - min_val)
            normalized[graph == 0] = 0
            if (j > 0) and (group_count[j - 1] < min_cells):
                normalized *= 0
            well_paints[j] = normalized

    painttttt = np.concatenate(well_paints, axis=1)
    fig = plt.figure(figsize=(4.5, 7.5))
    ax = fig.add_subplot(111)
    ax.get_xaxis().set_visible(False)
    ax.set_xticks([])
    ax.get_yaxis().set_visible(False)
    plt.imshow(painttttt, cmap="viridis", aspect='auto')
    if not hide_title:
        plt.title('spherocylindrical\nprojection', fontname='Arial', fontsize=24)
    for outline in optimized_outline:
        ax.plot(outline.T[1], outline.T[0], c="salmon", alpha=0.5)
    if savefig:
        fig.savefig(filename, bbox_inches='tight', dpi=160, transparent=True)
        plt.close(fig)


def _plate_triplot(folder, channel, canvas, z_cutoff=4, intensity_cutoff=50):
    """
    :param folder:
    :return:
    """

    # import data
    if not os.path.isfile(folder + 'accepted_cells.pk'):
        raise ValueError('Files not found, is {} an empty folder?'.format(folder))
    cells = pk.load(open(folder + 'accepted_cells.pk', 'rb'))
    data_dict = pk.load(open(folder + 'all_cells_miscellaneous_data.pk', 'rb'))
    morph_df = pd.read_excel(folder + 'all_cells_accepted_cells_morphology.xls')
    filtered_morph_df = morph_df[(morph_df['Quality_by_shape_index'] == 1) & (morph_df['is_outlier'] == 0)].copy()
    length_sorted_index = filtered_morph_df.sort_values(by='length').index

    # fig1: demograph
    demograph = data_dict[channel]['padded_axial_data'][length_sorted_index]
    # enhance contrast
    demograph[demograph > 0] = demograph[demograph > 0] + 0.1

    fig1 = plt.figure(figsize=(7.3, 4.5))
    ax = plt.subplot(111)
    ax.imshow(demograph.T, aspect='auto')
    ax.get_xaxis().set_visible(False)
    ticks = [0, 127, 255, 383, 511]
    tick_labels = [6, 3, 0, 3, 6]
    ax.set_yticks(ticks)
    ax.set_yticklabels(tick_labels, fontname='Arial', fontsize=16)
    ax.set_ylabel('distance from center [µm]', fontname='Arial', fontsize=16)
    #plt.title('normalized demograph', fontname='Arial', fontsize=16)
    fig1.savefig(folder + 'demgraph.png', bbox_inches='tight', dpi=160, transparent=True)
    plt.close()

    # fig2: spherocylindrical projection
    sc_projection = folder + 'spherocylindrical_projection.png'
    project_by_percentile(length_sorted_index,
                          cells,
                          channel,
                          canvas, savefig=True, filename=sc_projection)

    # puncta density plot:

    padded, xt_list, yt_list, nxt_list, nyt_list, optimized_outline = canvas
    coords = []
    for outline in optimized_outline:
        x0 = outline.T[1].mean()
        y0 = outline.T[0].mean()
        l = outline.T[0].max() - outline.T[0].min()
        w = outline.T[1].max() - outline.T[1].min()
        coords.append([x0, y0, l, w])

    fig3 = plt.figure(figsize=(4.5, 7.5))
    ax = fig3.add_subplot(111)
    # ax.get_xaxis().set_visible(False)
    ax.set_xticks([])
    # ax.set_xlabel('cell',fontname='Arial', fontsize=16)
    ax.get_yaxis().set_visible(False)
    # plt.imshow(painttttt, cmap="viridis")
    bg = np.concatenate(padded, axis=1)
    bg *= 0
    ax.imshow(bg, cmap='Greys', aspect='auto')
    plt.xlim(0, bg.shape[1]-1)

    for outline in optimized_outline:
        ax.plot(outline.T[1], outline.T[0], c="salmon", alpha=1)

    grouped_index_fraction = divide_by_fraction(len(length_sorted_index), by=6)
    grouped_data = [[], [], [], [], []]

    for i in range(5):
        p1 = grouped_index_fraction[i]
        p2 = grouped_index_fraction[i + 1]
        index_list = length_sorted_index[p1:p2]
        x0, y0, l, w = coords[4-i]

        for idx in index_list:
            reorient = False
            cell = cells[idx]
            puncta = cell.fluorescent_puncta[channel]
            cell_data = cell.data[channel][cell.mask > 0]
            axial_data=cell.measurements.signal_measurements[channel]['axial']
            half_l = int(0.5*len(axial_data))
            if axial_data[:half_l].mean() < axial_data[half_l:].mean():
                reorient = True

            if len(puncta) > 0:
                puncta_intensity = puncta[:, 2]
                normalized_puncta = puncta_projection(cell, channel)[:, -3:]
                ycoords = (normalized_puncta[:, 0] + normalized_puncta[:, 1]) - 0.5
                if reorient:
                    ycoords = -ycoords
                ycoords = ycoords * l + y0

                xcoords = (normalized_puncta[:, 2]) * w + x0

                z_scores = (puncta_intensity - cell_data.mean()) / cell_data.std()

                puncta_intensity = np.reshape(puncta_intensity, -1, 1)
                ycoords = np.reshape(ycoords, -1, 1)
                xcoords = np.reshape(xcoords, -1, 1)

                compiled_data = np.array([xcoords, ycoords, puncta_intensity, z_scores]).T
                compiled_data = compiled_data[np.where(np.abs(normalized_puncta[:, 2]<=0.51))]
                if len(compiled_data) > 0:
                    grouped_data[i].append(compiled_data)

        grouped_data[i] = np.concatenate(grouped_data[i], axis=0)
        accepted = np.where((grouped_data[i][:, 0] > 0) & (grouped_data[i][:, 2] > intensity_cutoff)
                            & (grouped_data[i][:, 3] > z_cutoff))
        grouped_data[i] = grouped_data[i][accepted]


        xy = grouped_data[i][:, :2]
        """
        dist_matrix = np.sqrt(np.sum((xy[:, np.newaxis] - xy) ** 2, axis=2))
        dist_matrix[dist_matrix == 0] = np.inf
        dist_matrix = 1 / dist_matrix
        weighted_dist = np.sum(dist_matrix, axis=1)
        weighted_dist /= weighted_dist.max()
        # weighted_dist = (weighted_dist-weighted_dist.min())/(weighted_dist.max()-weighted_dist.min())
        colors = cm.get_cmap('viridis')(weighted_dist)
        """


        #intensities = grouped_data[i][:, 2] / grouped_data[i][:, 2].max()

    merged_data=np.concatenate(grouped_data, axis=0)
    if len(merged_data) > 0:
        z = merged_data[:, 3]
        z = (z-z.min())/(z.max()-z.min())
        colors = cm.get_cmap('viridis')(z)
        ax.scatter(merged_data[:, 0], merged_data[:, 1], color=colors, alpha=0.2, s=80)

    #plt.title('fluorescent puncta\n density plot', fontname='Arial', fontsize=24)
    fig3.savefig(folder + 'puncta_density_plot', bbox_inches='tight',
                 dpi=160, transparent=True)
    plt.close()


from .helper import *
from .measure import Measurement
from skimage import registration
from . import io


class Cell:

    def __init__(self,
                 image_obj,
                 cluster_label,
                 cell_label,
                 mask,
                 bbox,
                 regionprop=None):
        """
        As the name implies, the Cell object takes in particles from each Cluster, computes the corresponding
        contour line and topological skeleton (midline) as well as many other measurements. It also contains a
        set of filter functions which determines whether a cell is to be accepted as a single bacterium or not.

        :param image_obj: master Image object
        :param cluster_label: corresponding Cluster id
        :param cell_label: cell id
        :param mask: binary mask
        :param bbox: refined bounding box
        :param regionprop: skimage regionprops object, made from mask if not provided
        """

        if regionprop != None:
            self.regionprop = regionprop
        else:
            self.regionprop = measure.regionprops(label_image=mask)[0]

        self.mask = ndi.median_filter(mask, 3) # smooths the edge of the binary mask using a median kernel
        self.pixel_microns = image_obj.pixel_microns
        self.bbox = bbox
        self.mask_channel_name = image_obj.mask_channel_name

        """
        Inherit data from Image
        """
        self.data = {} # Cell image data
        self.image_stat = {} # inherit global stats from Image
        (x1, y1, x2, y2) = self.bbox
        for channel, data in image_obj.data.items():
            self.data[channel] = data[x1:x2, y1:y2].copy()
            background_median = image_obj.image_background_median[channel]
            foreground_median = image_obj.image_foreground_median[channel]
            self.image_stat[channel] = [background_median, foreground_median]
        self.sobel = image_obj.mask_sobel[x1:x2, y1:y2]
        self.frangi = image_obj.mask_frangi[x1:x2, y1:y2]
        self.shape_index = image_obj.mask_shape_indexed[x1:x2, y1:y2]
        self.config = image_obj.config

        self.cluster_label = cluster_label
        self.cell_label = cell_label
        self.image_label = None
        self.init_contour = [] # list of cellular contour(s)
        self.skeleton_coords = [] # list of morphological skeleton(s)
        self.skeleton = None # binary skeleton image
        self.optimized_contour = [] # subpixel contour(s)
        self._contour_optimization_converged = None # whether or not the contour optimization process converges
        self.fluorescent_puncta = {} # dictionary of fluorescent puncta profile(s)
        self.midlines = [] # list of subpixel midlines
        self.width_lists = [] # width measures along midline
        self.measurements = Measurement() # initiate cellular Measurement
        self.profile_mesh = None
        self.contour_l1 = None
        self.contour_l2 = None
        self.widths_interp = None
        self.midline_interp = None

        """
        basic morphological metrics
        """
        self.largest_bend_angle = 0
        self.roughness = 1
        self.lengths = []
        self.vshaped = False
        self.branched = True
        self.discarded = False
        self.segmentation_quality_by_shapeindex = 1

        """
        test_mode
        """
        self.test_mode = bool(int(self.config['cell']['test_mode']))

    def find_contour(self):

        """
        masked gray scale image is created by replacing background (mask==0) pixels with foreground maxima (mask==1)
        the masked image is then smoothed using a gaussian filter
        incipient contour is generated using a marching square algorithm implemented in skimage.measure

        skimage.measure.find_contours occasionally return tandem duplicated contour, coordinates.
        this is probably introduced while closing the contour
        postprocessing inspection is therefore necessary
        """

        masked_phase = self.mask * self.data[self.mask_channel_name]
        masked_phase[self.mask == 0] = int(masked_phase.max())
        masked_phase = filters.gaussian(masked_phase, sigma=1)
        lvl = np.percentile(masked_phase[self.mask > 0], 98)
        contour_simplification_factor = float(self.config['cell']['contour_simplification_factor'])

        # only the outer contour enclosing the entire object is considered
        contour = measure.find_contours(masked_phase, level=lvl)[0]

        # find contour closing point
        x0, y0 = contour[0]
        closing_coords = np.where((contour[:, 0] == x0) & (contour[:, 1] == y0))[0]

        # in case the first two points are identical
        if len(closing_coords) == 1:
            contour = np.concatenate([contour,[contour[0]]])
        else:
            if closing_coords[1] == 1:
                p1, p2 = closing_coords[1], closing_coords[2]+1
                contour = contour[p1:p2]
            else:
                p1, p2 = closing_coords[0], closing_coords[1]+1
                contour = contour[p1:p2]

        # note that the recorded contour is by default closed
        #contour_length = measure_length(contour, pixel_microns=1)
        self.init_contour = spline_approximation(contour,
                                                 n=int(contour_simplification_factor*len(contour)),
                                                 smooth_factor=5)
        self.mask = draw.polygon2mask(self.mask.shape, self.init_contour)

    def extract_skeleton(self):

        """
        Extracts the pixelated topological skeleton of the binary mask
        note that the skimage embedded skeletonize function is sensitive
        to small extrusions near the edge. Although the binary mask was smoothed
        using a binary median filter, it is often not sufficient to suppress aberrant
        skeleton formation.

        Similar to MicrobeJ, OMEGA treats branched skeleton as a connected planar graph.
        The current implementation is suboptimal, as the graph is recorded using a
        default dictionary structure. For future versions, the Extract_skeleton function
        will be reconstructed with a NetworkX embedding.

        Cells with a severely bent skeleton is labeled as v-shaped. Notably, mycobacterial
        cell division is often succeeded by an abrupt force at the division site which
        immediately bends the divided cell and creates an angle of various degrees between the
        two sibling cells. Here we used a fixed cutoff (default value set to 50 degrees)
        to determine whether a particle is V-shaped, therefore a candidate for a microcluster
        containing two newly separated bacteria. The v-shaped particles are further segmented
        if the Cluster.split_branches function is being called.

        """

        if not self.discarded:
            self.skeleton_coords, self.skeleton = skeleton_analysis(self.mask)
            if len(self.skeleton_coords) == 1:
                xcoords = self.skeleton_coords[0][1]
                ycoords = self.skeleton_coords[0][2]

                # bend_angle function creates a moving window by the size of 2 * window.
                # skeleton shorter than 12 pixels (which should be discarded in most cases)
                # can not be properly handled.

                if len(xcoords) >= 12:
                    skel_coords = np.array([xcoords, ycoords]).T
                    self.largest_bend_angle = np.abs(bend_angle_open(skel_coords, window=5)).max()
                    if self.largest_bend_angle < float(self.config['cell']['largest_bend_angle']):
                        self.branched = False
                    else:
                        self.vshaped = True
            elif len(self.skeleton_coords) == 0:
                skel_length = np.count_nonzero(self.skeleton)
                if skel_length <= 3 or skel_length >= 1000:
                    self.discarded = True
                    if self.test_mode:
                        print('{}_{}_no skeleton found!'.format(self.cluster_label, self.cell_label))

    def optimize_contour(self):
        """
        infer object contour with subpixel resolution
        """
        optimize = bool(int(self.config['cell']['optimize_contour']))
        contour_enlarge_factor = float(self.config['cell']['contour_enlarge_factor'])
        expanded_contour = expand_contour(self.init_contour, scale=0.3)

        if not optimize:
            """
            use the initial contour rendered by skimage.measure.find_contours without further optimization    
            """
            self.optimized_contour = spline_approximation(self.init_contour,
                                                          n=int(2*len(self.init_contour)),
                                                          smooth_factor=2)
            self._contour_optimization_converged = True
        else:
            self.optimized_contour,\
            self._contour_optimization_converged = contour_optimization(expanded_contour,
                                                                        self.sobel)

        self.measurements.particle_morphology(self)

    def generate_midline_no_branch(self, fast_mode=False):
        """
        Our current midline algorithm doesn't work very well for branched objects
        future versions may include a graph embedded midline system that accounts for
        the various shapes
        :param fast_mode: extend the termini of the morphological skeleton to the poles and call it a day

        """
        if not self.branched and not self.discarded:
            pole1, pole2 = self.skeleton_coords[0][0]
            xcoords = self.skeleton_coords[0][1]
            ycoords = self.skeleton_coords[0][2]
            if len(xcoords) <= 10:
                if self.test_mode:
                    print('{}_{}_skeleton is too short!'.format(self.cluster_label,
                                                                self.cell_label))
                self.discarded = True
            else:
                if len(xcoords) <= 20:
                    interpolation_factor = 1
                else:
                    interpolation_factor = 2
                skel_coords = np.array([xcoords, ycoords]).T
                #
                smooth_skel = spline_approximation(skel_coords,
                                                   n=int(len(skel_coords)/interpolation_factor),
                                                   smooth_factor=5, closed=False)
                if not fast_mode:
                    smooth_skel, _converged = midline_approximation(smooth_skel,
                                                                    self.optimized_contour)
                    midline = extend_skeleton(smooth_skel, self.optimized_contour,
                                              find_pole1=pole1,
                                              find_pole2=pole2,
                                              interpolation_factor=interpolation_factor)

                else:
                    midline = extend_skeleton(smooth_skel, self.optimized_contour,
                                              find_pole1=pole1, find_pole2=pole2,
                                              interpolation_factor=1)

                width = direct_intersect_distance(midline, self.optimized_contour)
                length = line_length(midline)
                self.lengths.append(length*self.pixel_microns)
                self.midlines.append(midline)
                self.width_lists.append(width)

    def morphological_filter(self):
        """
        primary Cell object filter function using particle morphological descriptors. Empirically speaking,
        this function alone is sufficient to remove most of the falsely segmented objects, therefore no secondary
        vetting step is employed by the current edition of OMEGA, although an additional trained classifier could
        in theory further increase the segmentation accuracy.
        :return: None
        """
        morph_filter_keys = ['area', 'eccentricity',
                             'aspect_ratio',
                             'solidity',
                             'circularity',
                             'convexity',
                             'average_bending_energy',
                             'normalized_bending_energy',
                             'min_curvature',
                             'rough_Length',
                             'rough_sinuosity']
        like_cell = True
        for key in morph_filter_keys:
            val = self.measurements.morphology_measurements[key]
            low_threshold_config_key = key.lower()+'_low'
            high_threshold_config_key = key.lower()+'_high'
            low_threshold = float(self.config['cell'][low_threshold_config_key])
            high_threshold = float(self.config['cell'][high_threshold_config_key])
            like_cell *= in_range(val, low_threshold, high_threshold)

        if self.test_mode:
            print(self.vshaped, like_cell)
        if not like_cell:
            self.branched = True

    def generate_midline_branch_aware(self):
        # future function
        print('Function currently unavailable')

    def generate_measurements(self):
        """
        Signal and morphological profiling of Cell objects.
        :return: None
        """
        # does not support branched cell measurements yet
        # will include graph based branch analysis shortly

        if not self.branched and not self.discarded:
            if self.midlines == [] or self.width_lists == []:
                if self.test_mode:
                    print('No contour/midline coordinates found for cell {}-{}'.format(self.cluster_label,
                                                                                       self.cell_label))
                self.discarded = True
            else:
                correct_drift = bool(int(self.config['cell']['correct_cell_drift']))
                if correct_drift:
                    self._correct_xy_drift()
                try:
                    # signal profiling
                    unit_micron = float(self.config['cell']['profiling_unit_pixel'])
                    expanded_contour = expand_contour(self.optimized_contour, scale=0.5)
                    l1, l2, profiling_mesh,\
                    midline_interp = straighten_by_orthogonal_lines(expanded_contour,
                                                                    self.midlines[0],
                                                                    self.lengths[0],
                                                                    np.array(self.width_lists[0])*self.pixel_microns,
                                                                    unit_micron=unit_micron)

                    self.contour_l1 = l1
                    self.contour_l2 = l2
                    self.profile_mesh = profiling_mesh
                    self.midline_interp = midline_interp
                    self.widths_interp = np.sqrt(np.sum(np.square(l2-l1), axis=1))*self.pixel_microns
                    self.widths_interp = np.concatenate([[0], self.widths_interp, [0]])
                    self.measurements.signal(self, 'Shape_index', profiling_mesh)
                    for channel, img in self.data.items():
                        self.measurements.signal(self, channel, profiling_mesh)
                        if channel != self.mask_channel_name:
                            self.fluorescent_puncta[channel] = find_puncta(img, self.mask)

                    # morphological profiling
                    self.measurements.cell_morphology(self)
                except:
                    if self.test_mode:
                        print('{}_{}_contour optimization failed!'.format(self.cluster_label,
                                                                          self.cell_label))
                    self.discarded = True

    def compiled_cell_process(self):
        """
        Bundle of Cell functions
        :return:
        """
        self.optimize_contour()
        self.morphological_filter()
        self.generate_midline_no_branch()
        self.generate_measurements()
        # drop redundant data
        self.frangi = None
        self.sobel = None
        self.skeleton = None

    def _correct_xy_drift(self):
        """
        XY-drift correction at cellular level, may introduce numeric instability
        :return:
        """
        if not self.branched and not self.discarded:
            # invert phase contrast data

            ref_img = self.data[self.mask_channel_name]
            ref_img = 100+ref_img.max()-ref_img

            for channel, img in self.data.items():
                if channel != self.mask_channel_name:
                    shift, error, _diff = registration.phase_cross_correlation(ref_img, img,
                                                                               upsample_factor=10)
                    if np.max(np.abs(shift)) <= 3:
                        self.data[channel] = shift_image(img, shift)
                        if len(self.fluorescent_puncta) != 0:
                            puncta_info = self.fluorescent_puncta[channel]
                            puncta_info[:, 0] += shift[0]
                            puncta_info[:, 1] += shift[1]
                            self.fluorescent_puncta[channel] = puncta_info

    def _update_signal(self, channel):
        """
        post analysis signal update
        :param channel: data channel
        :return: None
        """
        midline = self.midlines[0]
        width = self.width_lists[0]
        mask = self.mask
        length = self.lengths[0]
        img = self.data[channel]
        self.measurements.update_signal(img, channel, midline, width, mask, length, self.profile_mesh)




import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import glob, os
from matplotlib import cm


def feature_plot(h, k, file_header, dst_folder, savefile=True, showplot=False):
    from matplotlib.gridspec import GridSpec as gs
    if k % 5 == 0:
        nrows = int(k / 5)
    else:
        nrows = int(k / 5) + 1
    ncols = 5
    grids = gs(nrows, ncols, hspace=0.5)
    fig = plt.figure(figsize=(2 * ncols, nrows*1.2))

    for i in range(k):
        x = int(i / 5)
        y = i % 5
        ax = fig.add_subplot(grids[x, y])
        ax.imshow(h[i].reshape(15, 30), aspect='auto', cmap='viridis')
        ax.set_title('feature {}'.format(i + 1))
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    if savefile:
        plt.savefig('{}{}_rank_{}.png'.format(dst_folder, file_header, k),
                    bbox_inches='tight', dpi=160, transparent=True)
    if not showplot:
        plt.close()




import configparser


class Preset_configurations_default:
    def __init__(self, configfile=None):
        self.config = configparser.ConfigParser()
        if configfile == None:
            self.config.read("./OMEGA_core/configurations/configuration_msm.ini")
            self.configurated = True
        else:
            self.config.read(configfile)
            self.configurated = True


from .helper import *
from .metrics import *
from .image import Image, _test_stitched_data
from .config import *
from .plot import *
import timeit
import glob, os, tifffile
import pickle as pk
from scipy import stats
import multiprocessing as mp
from matplotlib import rc
from matplotlib.gridspec import GridSpec
import warnings


"""
===================================
functions for automated processes
===================================
"""


def process_single_input(imagefile,
                         header,
                         dest_folder,
                         mask_channel_id=-1,
                         configfile='default',
                         print_time=False,
                         offset_correction=True,
                         split_branches=True,
                         save_input=False,
                         save_segmentation_result=False,
                         save_cells=True,
                         save_discarded=False,
                         save_branched=False,
                         boundary_classifier='./OMEGA_core/configurations/boundary_MLP.pk',
                         boundary_normalizer='./OMEGA_core/configurations/normalizer.pk',
                         return_obj=True):
    """

    :param imagefile: input image file
    :param header: user specified identification label
    :param dest_folder: path where the specified output dataset will be exported to
    :param mask_channel_id: index of the reference image
    :param configfile: user specified configuration file dest
    :param print_time: print time cost when done if set to True
    :param offset_correction: correct xy drift if set to True
    :param split_branches: extensive segmentation if set to True
    :param save_input: save processed images if set to True
    :param save_segmentation_result: plot and save segmentation result if set to True
    :param save_cells: pickle a list of cells if set to True
    :param save_discarded: include discarded particles when pickling if set to True
    :param save_branched: include branched particles when pickling if set to True
    :param boundary_classifier: path to pretrained model for boundary classification
    :param boundary_normalizer: path to pretrained model for boundary data normalization
    :param return_obj: return the processed Image object when done if set to True
    :return return the processed Image object if return_obj is True

    """
    config = None
    t1 = timeit.default_timer()
    if configfile == 'default':
        config = Preset_configurations_default().config
    elif '.ini' in configfile:
        try:
            config = Preset_configurations_default(configfile=configfile).config
        except:
            print('{} config file not supported, proceed with default settings.'.format(configfile))
            config = Preset_configurations_default().config

    # activate an Image object, load configurations
    obj = Image()
    obj.config = config

    # load image data
    if imagefile.endswith('.nd2'):
        obj.read_nd2_file(imagefile, mask_channel_id=mask_channel_id)
    elif imagefile.endswith('.tif'):
        obj.read_tiff_file(imagefile, mask_channel_id=mask_channel_id)
    else:
        raise ValueError('.{} file not supported!'.format(imagefile.split('.')[-1]))

    # inspect and calibrate input data
    obj.crop_edge(offset_correction=offset_correction)
    obj.enhance_brightfield()
    obj.enhance_fluorescence()

    # segmentation
    obj.locate_clusters()
    obj.cluster_segmentation(predictor=boundary_classifier,
                             normalizer=boundary_normalizer)
    obj.cell_segmentation(split_branches=split_branches)

    # retrieve cells
    cell_list = obj.get_cells(branched=save_branched,
                              discarded=save_discarded,
                              return_list=True)
    t2 = timeit.default_timer()
    if print_time:
        print("Analysis complete, {} seconds used.".format(round(t2 - t1), 2))
    if save_input:
        save_processed_images(obj, header, dest_folder)
    if save_segmentation_result:
        plot_segmented_image(obj, savefig=True, dpi=100, lw=0.3, scale=0.4,
                             show_branched=False,
                             show_accepted=True,
                             show_discarded=False,
                             filename="{}{}_segmentation_result.tif".format(dest_folder, header))
    if save_cells:
        pickle_dump(cell_list, header+'_cells', dest_folder)
    if return_obj:
        return obj


def process_folder(source_folder,
                   dest_folder=None,
                   file_type='.nd2',
                   mask_channel_id=-1,
                   configfile='default',
                   print_time=False,
                   offset_correction=True,
                   split_branches=True,
                   save_input=False,
                   save_segmentation_result=True,
                   save_cells=True,
                   save_discarded=False,
                   save_branched=False,
                   boundary_classifier='./OMEGA_core/configurations/boundary_MLP.pk',
                   boundary_normalizer='./OMEGA_core/configurations/normalizer.pk',
                   n_cores=2):
    """

    :param source_folder: folder where input image files are located
    :param dest_folder: path where the specified output dataset will be exported to
    :param file_type: image file type, could be either .tiff or .nd2
    :param mask_channel_id: index of the reference image
    :param configfile: user specified configuration file dest
    :param print_time: print time cost when done if set to True
    :param offset_correction: correct xy drift if set to True
    :param split_branches: extensive segmentation if set to True
    :param save_input: save processed images if set to True
    :param save_segmentation_result: plot and save segmentation result if set to True
    :param save_cells: pickle a list of cells if set to True
    :param save_discarded: include discarded particles when pickling if set to True
    :param save_branched: include branched particles when pickling if set to True
    :param boundary_classifier: path to pretrained model for boundary classification
    :param boundary_normalizer: path to pretrained model for boundary data normalization
    :param n_cores: number of CPU cores called in parallel

    """
    warnings.filterwarnings("ignore")

    # setup output folder and locate image data
    if dest_folder is None:
        dest_folder = source_folder + 'output/'
    if not os.path.isdir(dest_folder):
        os.mkdir(dest_folder)
    input_files = sorted(glob.glob(source_folder+'*'+file_type))
    if len(input_files) == 0:
        raise ValueError('No {} file(s) found in {}!'.format(file_type, source_folder))

    # multiprocessing
    for i in range(0, len(input_files), n_cores):
        processes = []
        for j in range(0, min(n_cores, len(input_files)-i)):
            file = input_files[i+j]
            header = file.split('/')[-1].split('.')[0]
            processes.append(mp.Process(target=process_single_input,
                                        args=(file, header, dest_folder,
                                              mask_channel_id,
                                              configfile,
                                              print_time,
                                              offset_correction,
                                              split_branches,
                                              save_input,
                                              save_segmentation_result,
                                              save_cells,
                                              save_discarded,
                                              save_branched,
                                              boundary_classifier,
                                              boundary_normalizer,
                                              False)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()


"""
========================================
functions for data import and export
========================================
"""


def pickle_dump(data, header, dest_folder):
    """
    pickle wrapper
    :param data: data of interest
    :param header: user specified file header/name
    :param dest_folder: destination folder
    :return: None
    """
    pk.dump(data, open("{}{}.pk".format(dest_folder, header), 'wb'))


def pickle_load_all_cells(folder):
    """
    load pickled cells from any folder
    :param folder: folder where *cells.pk files are stored
    :return: a list of retrieved Cell objects
    """
    pks = sorted(glob.glob(folder+'*cells.pk'))
    cells = []
    for file in pks:
        cells += pk.load(open(file, 'rb'))
    return cells


def save_processed_images(Image_obj, header, dest_folder):
    """
    save processed image data as 16-bit tif files
    :param obj: processed Image object
    :param header: user specified header/name
    :param dest_folder: destination folder
    :return: None
    """
    for channel, img in Image_obj.data.items():
        channel_name = channel.replace(' ', '_')
        tifffile.imsave("{}{}_{}_filtered.tif".format(dest_folder, header, channel_name),
                        img.astype(np.uint16), imagej=True)


def save_experiment(Image_obj, header, dest_folder):
    """
    save the entire Image object
    :param Image_obj: processed Image object
    :param header: user specified header/name
    :param dest_folder: destination folder
    :return: None
    """
    pickle_dump(Image_obj, header, dest_folder)


def get_measurements(cells,
                     header,
                     dest_folder):
    """
    save numerical measurements to excel sheets
    :param cells: list of cells
    :param header:
    :param dest_folder:
    :return:
    """
    branched_cells = []
    discarded_cells = []
    accepted_cells = []
    filtered_cell_dfs = {}
    outlier_id = []

    for i in range(len(cells)):
        cell = cells[i]
        if cell.discarded:
            discarded_cells.append(cell)
        elif cell.branched:
            branched_cells.append(cell)
        else:
            accepted_cells.append(cell)

    if discarded_cells is not []:
        discarded_df_dict = _to_dataframe(discarded_cells, by='particle_morphology')
        for key, df in discarded_df_dict.items():
            df.to_excel("{}{}_{}_{}.xls".format(dest_folder, header, 'discarded_cells', key))

    if branched_cells is not []:
        branched_df_dict = _to_dataframe(branched_cells, by='particle_morphology')
        for key, df in branched_df_dict.items():
            df.to_excel("{}{}_{}_{}.xls".format(dest_folder, header, 'branched_cells', key))

    if accepted_cells is not []:
        accepted_df_dict = _to_dataframe(accepted_cells, by='all')
        for key, df in accepted_df_dict.items():
            if key == 'morphology':
                filtered_df, outliers = _filter_outliers(df,
                                                         by=['length', 'width_median',
                                                             'width_std', 'circularity'])
            else:
                filtered_df, outliers = _filter_outliers(df,
                                                         by=['median'])
            filtered_df.to_excel("{}{}_{}_{}.xls".format(dest_folder, header, 'accepted_cells', key))
            filtered_cell_dfs[key] = filtered_df
            outlier_id.append(outliers)

    is_outlier = (np.sum(np.array(outlier_id), axis=0) > 0)*1
    return discarded_cells, branched_cells, accepted_cells, filtered_cell_dfs, is_outlier


def measure_by_folder(header,
                      dest_folder):
    """
    gather and export signal profiles
    :param dest_folder:
    :return:
    """

    cells = pickle_load_all_cells(dest_folder)
    discarded_cells, branched_cells, accepted_cells, \
    accepted_cell_dfs, outlier_id = get_measurements(cells, header, dest_folder)
    _miscellaneous = _export_data(accepted_cells,
                                  accepted_cell_dfs,
                                  outlier_id,
                                  header,
                                  dest_folder)

def _to_dataframe(cells,
                  by='all'):
    """
    convert cellular data to pandas dataframes
    :param cells: list of cells
    :param by: metrices of interest
    :return: dictionary of dataframes
    """
    all_keys = {}
    all_measurements = {}
    dataframe_dict = {}
    for cell in cells:
        image_label = cell.image_label
        cluster_label = cell.cluster_label
        cell_label = cell.cell_label
        key_dict, measurement_dict = cell.measurements.data_to_dict(by=by)
        for key, val in measurement_dict.items():
            keys = key_dict[key]
            if key not in all_keys:
                all_keys[key] = ['image_label', 'cluster_label', 'cell_label'] + keys
            if key not in all_measurements:
                all_measurements[key] = [[image_label, cluster_label, cell_label] + val]
            else:
                all_measurements[key].append([image_label, cluster_label, cell_label] + val)

    for key, matrix in all_measurements.items():
        columns = all_keys[key]
        dataframe_dict[key] = pd.DataFrame(matrix, columns=columns)
    return dataframe_dict


def _filter_outliers(df,
                     z_score_th=3,
                     iqr_th=2,
                     by=('length', 'width_median', 'width_std', 'circularity')):
    """
    combine outliers found by the z-score method of IQR method
    :param df: input multi-metrics dataframe
    :param z_score_th: z-score high bound
    :param iqr_th: IQR high bound
    :param by: metric keys of interest
    :return: updated dataframe, outlier identification list
    """

    # find outliers by zscore
    outlier_by_zscore = np.sum(stats.zscore(df[by], axis=0) >= z_score_th, axis=1) >= 1

    # find outliers by IQR
    q1,q3 = df[by].quantile(0.25), df[by].quantile(0.75)
    iqr = q3-q1
    outlier_by_iqr = np.sum(np.array((df[by] < q1-iqr_th*iqr) | (df[by] > q3+iqr_th*iqr)),
                            axis=1) >= 1
    df['is_outlier'] = ((outlier_by_iqr + outlier_by_zscore) > 0)*1
    return df, list(df['is_outlier'])


def _export_data(cell_list,
                 cell_dataframe_dict,
                 is_outlier,
                 header,
                 dest_folder,
                 save_data=True):

    cell_profiles = {}
    if 'morphology' not in cell_dataframe_dict:
        raise ValueError('Morphological dataframe not found!')

    morph_df = cell_dataframe_dict['morphology']
    if 'is_outlier' not in morph_df.columns:
        raise ValueError('Outlier information not found!')

    # record cell lengths and authenticity
    cell_profiles['is_outlier'] = is_outlier
    cell_profiles['lengths'] = morph_df['length'].values
    cell_profiles['cell_id'] = []
    counter = 0
    for key, df in cell_dataframe_dict.items():
        if key not in ['morphology', 'Shape_index']:

            axial_padded = []
            axial_normalized = []
            axial_straighten = []
            axial_straighten_width_normalized = []
            lateral_normalized = []
            channel_dict = {}
            for cell in cell_list:
                if counter == 0:
                    cell_profiles['cell_id'].append([cell.cluster_label, cell.cell_label])
                axial_data = cell.measurements.signal_measurements[key]['axial']
                axial_data_normalized = normalize_data1D(axial_data, re_orient=True, base=5)
                lateral_data = cell.measurements.signal_measurements[key]['lateral_mean']
                lateral_data_normalized = normalize_data1D(lateral_data, re_orient=False, base=5)
                axial_normalized.append(list(np.interp(np.linspace(0, 1, 100),
                                                       np.linspace(0, 1, len(axial_data_normalized)),
                                                       axial_data_normalized)))
                lateral_normalized.append(list(np.interp(np.linspace(0, 1, 100),
                                                         np.linspace(0, 1, len(lateral_data_normalized)),
                                                         lateral_data_normalized)))

                # pad_data, necessary for plotting demograph
                length = cell.measurements.morphology_measurements['length']
                axial_padded.append(list(pad_data(axial_data_normalized, length,
                                                  normalize=False, max_len=12)))

                # straightened data, necessary for making spherocylindrical projection plot
                straightened_data = cell.measurements.signal_measurements[key]['straighten']
                straightened_data = normalize_data_2D(straightened_data,
                                                      percentile_low_bound=0,
                                                      re_orient=True)
                axial_straighten.append(straightened_data)

                #
                width_normalized = cell.measurements.signal_measurements[key]['straighten_normalized']
                width_normalized = normalize_data_2D(width_normalized,
                                                     percentile_low_bound=0,
                                                     re_orient=True)
                axial_straighten_width_normalized.append(width_normalized)
            process_folder
            channel_dict['padded_axial_data'] = np.array(axial_padded)
            channel_dict['length_normalized_axial_data'] = np.array(axial_normalized)
            channel_dict['width_normalized_lateral_data'] = np.array(lateral_normalized)
            channel_dict['straightened_axial_data'] = axial_straighten
            channel_dict['straightened_width_normalized_axial_data'] = axial_straighten_width_normalized
            cell_profiles[key] = channel_dict
            counter = 1
    if save_data:
        pk.dump(cell_profiles, open('{}{}_miscellaneous_data.pk'.format(dest_folder, header), 'wb'))
    return cell_profiles



from .helper import *


class Boundary:

    def __init__(self, p1, p2, b_coords):
        self.p1 = p1
        self.p2 = p2
        self.boundary_coordinates = b_coords
        self.false_boundary = False
        self.relative_distance = 0.0
        self.deflection_angle = 1.0
        self.center_line_angle = 1.0
        self.bbox = []
        self.length = 0
        self.max_perimeter_fraction = 0.0
        self.boundary_average_intensity = 0.0
        self.phase_intensity_deviation = 0.0
        self.shapeindex_median = 0
        self.metrics = {}

    def primary_boundary_metrics(self, cluster):

        self.length = len(self.boundary_coordinates)
        self.length = estimate_boundary_length(np.array(self.boundary_coordinates))

        # retrieve particle regionprops
        P1, P2 = cluster.particles[self.p1],\
                 cluster.particles[self.p2]

        # centroid coordinates
        x1, y1 = P1.centroid
        x2, y2 = P2.centroid

        # particle bbox
        bbox1 = P1.bbox
        bbox2 = P2.bbox
        self.bbox = [min(bbox1[0], bbox2[0]), min(bbox1[1], bbox2[1]),
                     max(bbox1[2], bbox2[2]), max(bbox1[3], bbox2[3])]

        # particle major axis length
        L1 = P1.major_axis_length
        L2 = P2.major_axis_length

        # areas of the two particles
        A1, A2 = P1.area, P2.area

        # mean intensity
        mean_intensity = (P1.mean_intensity*A1 + P2.mean_intensity*A2)/(A1+A2)

        # perimeters of two particles:
        per1, per2 = P1.perimeter, P2.perimeter

        # calculate particle orientation
        theta1, theta2 = orientation_by_eig(P1), orientation_by_eig(P2)

        # orientation of line connecting two centroids
        thetaP1P2 = np.arctan((x2 - x1) / (y2 - y1))

        # xy coordinates of boundary
        boundary_x, boundary_y = np.array(self.boundary_coordinates).T

        # euclidean distance between two centroids
        dP1P2 = distance([x1 + bbox1[0], y1 + bbox1[1]], [x2 + bbox2[0], y2 + bbox2[1]])

        # if distance between two centroids approximates a half of total lengths,
        # the two objects are more likely to be from one cell
        self.relative_distance = 2 * dP1P2 / (L1 + L2)
        if self.relative_distance > 1:
            self.relative_distance == 1

        # deflection angle between two objects, the smaller it is , the more likely
        # the two particles belong to one
        self.deflection_angle = deflection_angle(theta1, theta2) / (0.5 * np.pi)
        self.center_line_angle = max(deflection_angle(theta1, thetaP1P2),
                                     deflection_angle(theta1, thetaP1P2)) / (0.5 * np.pi)

        # if the length of the boundary is larger than over half of the length of the smaller object,
        # the two objects are more likely to be from one cell
        self.max_perimeter_fraction = self.length / min(per1, per2)

        # signal intensities along boundary
        self.boundary_average_intensity = np.mean(cluster.phase[boundary_x, boundary_y].flatten())
        self.boundary_sobel_mean = cluster.sobel[boundary_x, boundary_y].mean()
        self.boundary_sobel_std = cluster.sobel[boundary_x, boundary_y].std()
        self.boundary_normalized_intensity = self.boundary_average_intensity/cluster.global_mask_foreground_median
        self.phase_intensity_deviation = abs(self.boundary_average_intensity-mean_intensity) / \
                                         cluster.global_mask_background_std

        self.shapeindex_std = cluster.shape_indexed[boundary_x, boundary_y].std()
        self.shapeindex_mean = cluster.shape_indexed[boundary_x, boundary_y].mean()
        self.frangi_mean = cluster.frangi[boundary_x, boundary_y].mean()
        self.frangi_std = cluster.frangi[boundary_x, boundary_y].std()
        self.DoG_mean = cluster.DoG[boundary_x, boundary_y].mean()
        self.DoG_std = cluster.DoG[boundary_x, boundary_y].std()
        self.RoG_mean = cluster.RoG[boundary_x, boundary_y].mean()
        self.RoG_std = cluster.RoG[boundary_x, boundary_y].std()
        self.min_neighbor_size = (min(A1, A2))*(cluster.pixel_microns**2)

        self.metrics = {'Maximum_perimeter_fraction': self.max_perimeter_fraction,
                        'Min_neighbor_size': self.min_neighbor_size,
                        'Boundary_length': self.length,
                        'Particle_deflection_angle': self.deflection_angle,
                        'Centroid_deflection_angle': self.center_line_angle,
                        'Normalized_centroid_distance': self.relative_distance,
                        'Intensity': self.boundary_average_intensity,
                        'Intensity_deviation': self.phase_intensity_deviation,
                        'Shapeindex_mean': self.shapeindex_mean,
                        'Shapeindex_variation': self.shapeindex_std,
                        'Sobel_mean': self.boundary_sobel_mean,
                        'Sobel_variation': self.boundary_sobel_std,
                        'Frangi_mean': self.frangi_mean,
                        'Frangi_std': self.frangi_std,
                        'DoG_mean': self.DoG_mean,
                        'DoG_std': self.DoG_std,
                        'RoG_mean': self.RoG_mean,
                        'RoG_std': self.RoG_std}

    """
    def boundary_classification_default(self, cluster):

        min_boundary_length = int(float(cluster.config['width_low']) / (cluster.pixel_microns * np.sqrt(2)))

        if self.max_perimeter_fraction > float(cluster.config['perimeter_fraction_cutoff']):
            self.false_boundary = True

        elif self.length >= min_boundary_length:

            byShape = bool(max(self.deflection_angle, self.center_line_angle) <\
                           float(cluster.config['max_deflection_angle']))

            byDist = bool(self.relative_distance >= float(cluster.config['normalized_centroid_distance_min']))

            byShapeid = bool(self.shapeindex_mean <= float(cluster.config['max_boundary_shape_index']))

            byFrangi = bool(self.frangi_mean >= float(cluster.config['min_boundary_frangi']))

            if np.sum([byShape, byDist, byShapeid, byFrangi]) >= 3:
                self.false_boundary = True


    
    deprecated functions
    def boundary_classification_SVM(self, predictor):

        self.false_boundary = False

    def boundary_classification_decision_tree(self, predictor):

        boundary_data = np.array(list(self.metrics.values())).reshape(1, -1)
        self.false_boundary = bool(predictor.predict(boundary_data)[0])

    def boundary_visualization(self, cluster):
        x1,y1,x2,y2 = self.bbox
        phase_img = cluster.phase[x1:x2,y1:y2]
        mask_img = (cluster.watersheded == self.p1)[x1:x2,y1:y2] & (cluster.watersheded == self.p2)[x1:x2, y1:y2]
        _xcoords, _ycoords = np.array(self.boundary_coordinates).T
        xcoords = _xcoords-x1
        ycoords = _ycoords-y1
        return phase_img, mask_img, xcoords, ycoords
    """

def boundary_neighbor_pairwise(watershed, boundary_mask, connectivity=1):
    particle_pairwise_dict = {}
    boundary_pairwise_dict = {}
    xlist, ylist = np.where(boundary_mask > 0)
    xmax, ymax = watershed.shape
    n = 0
    unique_neighbors = []
    for i in range(len(xlist)):
        x, y = xlist[i], ylist[i]
        # find unique neighbors
        if connectivity == 1:
            m = find_unique([watershed[max(x - 1, 0), y], watershed[min(x + 1, xmax - 1), y],
                             watershed[x, max(y - 1, 0)], watershed[x, min(ymax - 1, y + 1)]])
        elif connectivity == 2:
            m = find_unique(watershed[max(x - 1, 0):min(x + 1, xmax - 1), max(y - 1, 0):max(y - 1, 0)].flatten())

        # discard background pixels
        if (m[0] == 0) & (len(m) >= 3):
            unique_neighbors = m[1:]
        elif (m[0] != 0) & (len(m) >= 2):
            unique_neighbors = m
        if len(unique_neighbors) != 0:
            for pair in combinations(unique_neighbors, 2):
                p1, p2 = pair
                if p1 not in particle_pairwise_dict:
                    particle_pairwise_dict[p1] = {p2: n}
                    n += 1
                elif p2 not in particle_pairwise_dict[p1]:
                    particle_pairwise_dict[p1][p2] = n
                    n += 1
                idx = particle_pairwise_dict[p1][p2]
                if idx not in boundary_pairwise_dict:
                    boundary_pairwise_dict[idx] = Boundary(p1,p2,[[x,y]])
                else:
                    boundary_pairwise_dict[idx].boundary_coordinates += [[x, y]]
    return particle_pairwise_dict, boundary_pairwise_dict


def find_unique(datalist):
    return sorted(list(set(datalist)))


def deflection_angle(theta1, theta2):
    return min(abs(theta1 + theta2), abs(theta1 - theta2))


def boundary_annotation_thumbnail(boundary, cluster):
    x1,y1,x2,y2 = boundary.bbox
    phase = np.round(normalize_data1D(cluster.phase[x1:x2,y1:y2],
                                      re_orient=False, base=0)*255).astype(np.uint8)
    coords = np.array(boundary.boundary_coordinates)
    phase_labeled = phase.copy()
    phase_labeled[coords[:,0]-x1,coords[:,1]-y1] = 255
    return np.concatenate([phase,phase_labeled],axis=1)







from .helper import *
from scipy.stats import kurtosis, skew, linregress
from scipy import signal


def center_of_mass(data, xcor):
    center = np.sum((data) * xcor) / np.sum((data))
    return np.round(center, 3)


def normalize_array(data, min_val, max_val, base=0):
    return (data-min_val+base)/(max_val-min_val+base)


def weighted_segmentation(data, quantile_filter=0.05, weighted_center=True):
    if quantile_filter:
        min_threshold = np.quantile(data, quantile_filter)
    else:
        min_threshold = 0
    filtered_data = data.copy()
    normalized_data = normalize_array(filtered_data, min_threshold, filtered_data.max())
    normalized_data[filtered_data <= min_threshold] = 0.0
    xcor = np.linspace(0, 1, len(normalized_data))
    if weighted_center:
        com_0 = center_of_mass(normalized_data, xcor)
    else:
        com_0 = 0.5
    adjusted_center_dist_1 = int(com_0*len(normalized_data))
    adjusted_center_dist_2 = len(normalized_data) - adjusted_center_dist_1
    first_half = normalized_data[:adjusted_center_dist_1]
    xcor_first = xcor[:adjusted_center_dist_1]
    second_half = normalized_data[-adjusted_center_dist_2:]
    xcor_second = xcor[-adjusted_center_dist_2:]
    com_1 = center_of_mass(first_half, xcor_first)
    com_2 = center_of_mass(second_half, xcor_second)
    com_dist = com_2-com_1
    return [com_1, com_0, com_2], round(com_dist, 3)


def segmented_mean(data, quantile_filter=0.05, weighted_center=True):
    interpolated = np.interp(np.linspace(0, 1, 200), np.linspace(0, 1, len(data)), data)
    smoothed = moving_window_average(interpolated)
    [c1, c2, c3], c_dist = weighted_segmentation(smoothed,\
                                                 quantile_filter=quantile_filter,\
                                                 weighted_center=weighted_center)
    n_points = len(smoothed)
    c1, c2, c3 = int(c1*n_points), int(c2*n_points), int(c3*n_points)
    l1, l2, r1, r2 = smoothed[:c1].mean(), smoothed[c1:c2].mean(), smoothed[c2:c3].mean(), smoothed[c3:].mean()
    h1, h2 = smoothed[:c2].mean(), smoothed[c2:].mean()
    pole, center = np.concatenate([smoothed[:c1], smoothed[c3:]]).mean(), smoothed[c1:c3].mean()
    return l1, l2, r1, r2, h1, h2, pole, center


def sigmoid(x):
    return 1/(1+np.exp(-x))


def measure_symmetry(data, weighted):
    if weighted:
        weighted_center = center_of_mass(data, np.linspace(0, 1, len(data)))
    else:
        weighted_center = 0.5
    avg_1 = data[:int(round(weighted_center * len(data)))].sum()
    avg_2 = data[int(round(weighted_center * len(data))):].sum()
    ratio = max(avg_1, avg_2)/min(avg_1, avg_2)
    return 2-2*sigmoid(ratio-1)

"""
Deprecated
def measure_centrifugality(pole, center):
    return round(np.log2(pole/center), 3)


def measure_membrane_idx(data, quantile_filter=0.2, weighted_center=True, window=2):
    interpolated = np.interp(np.linspace(0, 1, 100), np.linspace(0, 1, len(data)), data)
    smoothed = helper.moving_window_average(interpolated)
    [c1, c2, c3], c_dist = weighted_segmentation(smoothed,\
                                                 quantile_filter=quantile_filter,\
                                                 weighted_center=weighted_center)
    n_points = len(smoothed)
    c1, c2, c3 = int(c1*n_points), int(c2*n_points), int(c3*n_points)
    v1 = smoothed[max(0, c1-window):min(len(smoothed), c1+window)].mean()
    v2 = smoothed[max(0, c2-window):min(len(smoothed), c2+window)].mean()
    v3 = smoothed[max(0, c3-window):min(len(smoothed), c3+window)].mean()
    ratio = round(np.average([v1, v3])/v2, 3)
    return ratio, c_dist
"""

def divide_cell(l, tip=0.25):
    left_tip_fraction = tip/l
    left_subpolar_fraction = tip/l+min(0.1, 1.5/l)
    left_periseptum_fraction = 0.4
    septum_fraction = 0.6
    right_perimseptum_fraction = 1-tip/l-min(0.1, 1.5/l)
    right_subpolar_fraction = 1-tip/l
    right_tip_fraction = 1
    return [left_tip_fraction, left_subpolar_fraction,\
            left_periseptum_fraction, septum_fraction,\
            right_perimseptum_fraction,
            right_subpolar_fraction, right_tip_fraction]


def divede_cell_pos(data, l, normalize=True):
    if normalize:
        data = normalize_data1D(data, re_orient=True)
    fractions = divide_cell(l)
    n_points = len(data)
    coords = np.round(np.array(fractions)*n_points).astype(int)
    output_data = []
    last_pos = 0
    for pos in coords:
        output_data.append(np.mean(data[last_pos:pos]))
        last_pos = pos
    return output_data


def _divede_cell_pos_raw(data, l):
    half_l = int(0.5 * len(data))
    if data[:half_l].mean() < data[-half_l:].mean():
        data = np.flip(data)
    fractions = divide_cell(l)
    n_points = len(data)
    coords = np.round(np.array(fractions)*n_points).astype(int)
    output_data = []
    last_pos = 0
    for pos in coords:
        output_data.append(np.mean(data[last_pos:pos]))
        last_pos = pos
    return output_data


def kurtosis_skewness(data):
    return kurtosis(data), skew(data)


def scaled_linregress(data_1D):
    data_1D = signal.medfilt(data_1D,kernel_size=5)
    slope, intercept, r_value, p_value, std_err = linregress(np.linspace(0,1,len(data_1D)),data_1D/data_1D.max())
    return slope, intercept, r_value, p_value, std_err


def FWHM_background_aware(data, percentile = False):
    confidence_score = 2
    FWHM = 0
    L = len(data)
    if percentile:
        max_val = np.percentile(data, 95)
        min_val = np.percentile(data, 5)
    else:
        max_val = data.max()
        min_val = data.min()
    HM = min_val + 0.5*(max_val-min_val)
    v1 = data[:-1]
    v2 = data[1:]
    p1 = v1-HM
    p2 = v2-HM
    p3 = p1*p2
    intersections = np.where(p3<0)[0]
    intersection_xcoords = []
    for i in intersections:
        intersection_xcoords.append(linear_intersection(i,v1[i],i+1,v2[i],HM))
    if len(intersection_xcoords) == 1:
        idx = np.where(data == data.max())[0]
        if min(L-idx,idx) >= 5:
            FWHM = abs(intersection_xcoords[0]-int(idx))*2
            confidence_score = 1
    if len(intersections) == 2:
        FWHM = intersections[1]-intersections[0]
        confidence_score = 0
    elif len(intersections) >= 3:
        FWHM = intersections[-1]-intersections[0]
        confidence_score = 1
    return FWHM, confidence_score

def linear_intersection(x0,y0,x1,y1,y):
    if y1 == y0:
        return 0.5*(x1+x0)
    else:
        return round(((x1-x0)*(y-y1)/(y1-y0))+x1,3)

# morphological metrics
def sinuosity(midline):
    if (midline[0]-midline[-1]).sum() == 0:
        raise ValueError('Midline coordinates appear to be closed!')
    end_to_end_dist = distance(midline[0],midline[-1])
    length = line_length(midline)
    ret = round(length/end_to_end_dist, 3)
    if ret < 1:
        ret = 1.0
    return ret


def average_bending_energy(contour, window=3):
    angles = bend_angle(contour[:-1], window=window)*np.pi/180
    ks = (np.sin(angles)**2).sum()/len(contour)
    return round(ks, 4)


def bending_energy(data, window=3):
    """
    Bending energy func by Bowie and Young, 1977
    BE = (1/P) * (sum(K(p)^2 * dp))
    P: contour length
    dp: unit contour arc length
    K(p): the rate of change in tangent direction θ of the contour,
          as a function of the arc length p (dp)
          K(p) = d_theta/dp
    :param data:
    :param window:
    :return:
    """
    p1 = np.concatenate((data[-window:], data[:-window])).T
    p2 = data.copy().T
    p3 = np.concatenate((data[window:], data[0:window])).T
    p1p2 = p1[0]*1+p1[1]*1j - (p2[0]*1+p2[1]*1j)
    p1p3 = p1[0]*1+p1[1]*1j - (p3[0]*1+p3[1]*1j)

    d_theta = np.angle(p1p3/p1p2)
    d_p = np.sqrt(np.sum(np.square(p2-p1).T, axis=1))
    P = d_p.sum()
    K_p = d_theta/d_p

    BE = np.sum(K_p**2/d_p)/P

    return BE


def circularity(regionprop):
    perimeter, area = regionprop.perimeter, regionprop.area
    return round(4*np.pi*area/(perimeter**2), 3)


def convexity(regionprop):
    image_perimeter = regionprop.perimeter
    convex_image_perimeter = measure.regionprops(regionprop.convex_image*1)[0].perimeter
    return round(convex_image_perimeter/image_perimeter, 3)


def width_omit_odds(width_list):
    new_list = []
    for w in width_list:
        new_list += list(w[2:-2])
    return np.array(new_list)


def angle_between_vector(v1, v2):
    d1 = np.sqrt(np.sum(v1 ** 2))
    d2 = np.sqrt(np.sum(v2 ** 2))
    return np.arccos(np.dot(v1, v2) / (d1 * d2))


def puncta_projection(cell, channel):
    puncta = cell.fluorescent_puncta[channel]
    p_puncta = puncta[:, :2]
    midline = cell.midlines[0]
    width = cell.width_lists[0]
    L = measure_length(midline)
    output = []
    for pp in p_puncta:
        v_p = pp[np.newaxis, :] - midline
        v_p_length = np.sqrt(np.sum(v_p ** 2, axis=1))

        nearest_loc = np.argmin(v_p_length)
        if nearest_loc == len(midline) - 1:
            v_m = midline[-1] - midline[-2]
        else:
            v_m = midline[nearest_loc + 1] - midline[nearest_loc]

        dl = v_p_length[nearest_loc]
        v_x_proj = np.array([1, 0])

        ang1 = angle_between_vector(v_m, v_x_proj)
        ang2 = angle_between_vector(v_p[nearest_loc], v_x_proj)
        ang12 = ang1 - ang2
        dx = np.cos(ang12) * dl
        dy = np.sin(ang12) * dl

        if nearest_loc == 0:
            l = 0
        else:
            l = measure_length(midline[:nearest_loc + 1])

        w = width[nearest_loc]

        if w == 0:
            norm_dy = dy / (2*np.max(width))
            norm_dx = dx / L
        else:
            norm_dx, norm_dy = dx / L, dy / w
        norm_l = l / L
        output.append([l, dl, dx, dy, norm_l, norm_dx, norm_dy])
    output = np.array(output)
    return output


def minimal_intrapuncta_distance(cell, channel):
    puncta = cell.fluorescent_puncta[channel]
    x_coords = puncta[:, 0]
    y_coords = puncta[:, 1]
    x_dist = x_coords[np.newaxis, :] - x_coords[:, np.newaxis]
    y_dist = y_coords[np.newaxis, :] - y_coords[:, np.newaxis]
    dist_matrix = np.sqrt(x_dist ** 2 + y_dist ** 2)
    dist_matrix[dist_matrix == 0] = np.inf
    return np.min(dist_matrix, axis=0) * cell.pixel_microns


def fractional_positions(cell, channel):
    counts = []
    fractions = np.array([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])
    if len(cell.fluorescent_puncta[channel]) > 0:
        projected = puncta_projection(cell, channel)
        x_pos = projected[:, 4] + projected[:, 5]
        pos_count = np.sum((x_pos[:, np.newaxis] - fractions) >= 0, axis=1) - 1
        pos_count[pos_count == -1] = 0
        pos_count[pos_count == 5] = 4
        for i in range(5):
            counts.append(np.count_nonzero(pos_count == i))
    else:
        counts = [0] * 5
    return (counts)


def puncta_midline_dist(cell, channel):
    if len(cell.fluorescent_puncta[channel]) > 0:
        midline = cell.midlines[0]
        puncta = cell.fluorescent_puncta[channel]
        xy_coords = puncta[:, :2]
        dist_matrix = np.sqrt(np.sum((xy_coords[:, np.newaxis] - midline) ** 2, axis=2))
        dist_min = np.min(dist_matrix, axis=1).mean()
    else:
        dist_min = 0
    return dist_min


def puncta_contour_dist(cell, channel):
    if len(cell.fluorescent_puncta[channel]) > 0:
        contour = cell.optimized_contour
        puncta = cell.fluorescent_puncta[channel]
        xy_coords = puncta[:, :2]
        dist_matrix = np.sqrt(np.sum((xy_coords[:, np.newaxis] - contour) ** 2, axis=2))
        dist_min = np.min(dist_matrix, axis=1).mean()
    else:
        dist_min = 0
    return dist_min


def symmetric_point(data1d):
    xcor = np.linspace(0, 1, 100)
    data = np.interp(xcor,
                     np.linspace(0, 1, len(data1d)),
                     data1d)
    cumsum = np.cumsum(data)

    half_sum = 0.5 * data.sum()
    i1, i2 = np.where(cumsum > half_sum)[0][0:2]
    v1, v2 = cumsum[i1], cumsum[i2]
    x1, x2 = xcor[i1], xcor[i2]

    x = ((half_sum - v1) / (v2 - v1)) * (x2 - x1) + x1

    if x >= 0.5:
        x = 1 - x
    return round(x, 3)


def normalized_contour_complexity(cell, window=1):
    contour = cell.init_contour
    angles = np.abs(bend_angle(contour, window=window))
    angles[angles <= 15] = 0
    complexity_cell = 0.5*angles.sum()/len(contour)
    area = cell.regionprop.area
    L = cell.regionprop.major_axis_length
    skeleton_L = len(np.nonzero(cell.skeleton)[0])
    L = max(L, skeleton_L)
    standard_complexity = standard_rod_complexity(area, L, window, 2*len(contour))
    return complexity_cell/standard_complexity


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA, NMF
import os, glob
import pickle as pk
from .plot import feature_plot


# now conduct NMF
def NMF_test(matrix,
             n_components=[20],
             rand_state=[47],
             alpha=[0.1],
             tol=[0.001],
             dst_folder=None,
             plot=True,
             save_plots=False,
             save_models=True):
    import pickle as pk
    # test whether input matrix has the right format
    if min(matrix.shape) <= 1:
        raise ValueError('For N x M matrix both N and M should be no smaller than 2.')
    if matrix.min() < 0:
        raise ValueError('Matrix should be non-negative.')
    if np.isnan(matrix).sum() > 0:
        raise ValueError('NaNs found in matrix')

    # create dst folder
    if dst_folder is not None:
        if not os.path.isdir(dst_folder):
            os.mkdir(dst_folder)
    parameter_dict = {}
    models = {}
    for n in n_components:
        for rd in rand_state:
            for a in alpha:
                for t in tol:
                    k = '{}_comp_{}_rdstate_{}_alpha_{}_tol'.format(n, rd, a, t)
                    parameter_dict[k] = [n, rd, a, t]
    for k, par in parameter_dict.items():
        n, rd, a, t = par
        model = NMF(init='nndsvdar',
                    shuffle='True',
                    n_components=n,
                    solver='cd',
                    random_state=rd,
                    alpha=a,
                    tol=t,
                    max_iter=2000)
        w = model.fit_transform(matrix)
        h = model.components_
        residual = model.reconstruction_err_
        if plot:
            feature_plot(h, n,
                         file_header=k,
                         dst_folder=dst_folder,
                         savefile=save_plots,
                         showplot=False)
        models[k] = [model, w, residual]
    if save_models:
        pk.dump(models, open('{}all_models.pk'.format(dst_folder), 'wb'))
    return models


__author__ = "jz-rolling"

from .helper import *
from .cell import Cell
from . import boundary
import skimage as sk
import numpy as np
from sklearn import svm,linear_model,tree
import pickle as pk


class Cluster:
    """
    A cluster is a continuous patch of the binary mask that contains one to many cell-like particles
    """
    def __init__(self, image_obj=None,
                 cluster_regionprop=None):
        """
        :param image_obj: corresponding master Image object
        :param cluster_regionprop: Skimage regionprop object of the corresponding Cluster

        Attributes
        ==========

        mask
        ---------- binary mask of the Cluster

        shape_indexed
        ---------- clip of shape index image

        labeled_seeds
        ---------- particle seeds labled by unique integers (1->N)

        particle_counts
        ---------- number of particles found in Cluster

        watershed
        ---------- watershed segmentation output

        particles
        ---------- dictionary of found particles

        boundary_mask
        ---------- binary image of inter-particle boundaries

        particle_pairwise_dict & boundary_pairwise_dict
        ---------- dictionaries recording each directly contacting particles and the assigned boundaries
                   For example: a particle_pairwise_dict entry may look like: {a : {b : c}}, here a,b,c are all
                   integers that denote the numeric labels of particle A, particle B and the boundary C between
                   A and B. A boundary_pairwise_dict entry may look like {c:Boundary_obj} where boundary C is
                   represented by a Boundary class.

        merge_particle_info
        ---------- conflated particle-boundary information

        cells
        ---------- dictionary of Cell object(s)

        discarded
        ---------- Cluster omitted from analysis if True

        touching_edge
        ---------- Cluster contains edge pixels if True

        bbox
        ---------- coordinates of the bounding box corners by the format of (x1, y1, x2, y2)
                   note that the bounding box is based on an inflated ROI
                   rather than the original ROI (regionprops.bbox)

        label
        ---------- integer label of the Cluster

        phase, sobel, mask_clumpy_zone
        ---------- clips of corresponding image data inherited from master Image object

        pixel_microns
        ---------- unit pixel length

        config
        ---------- Cluster configurations

        """
        self.mask = None
        self.shape_indexed = None
        self.labeled_seeds = None
        self.particle_counts = 0
        self.watershed = None
        self.particles = {}
        self.boundary_mask = None
        self.particle_pairwise_dict = {}
        self.boundary_pairwise_dict = {}
        self.merge_particle_info = None
        self.cells = {}
        self.discarded = False
        self.touching_edge = False
        self.false_boundary_neighbors = []
        self.false_boundary_id = []

        if (image_obj is not None) and (cluster_regionprop is not None):
            self.bbox = optimize_bbox(image_obj.shape, cluster_regionprop.bbox)
            (x1, y1, x2, y2) = self.bbox
            self.label = cluster_regionprop.label
            self.phase = image_obj.data[image_obj.mask_channel_name][x1:x2, y1:y2].copy()
            self.sobel = image_obj.mask_sobel[x1:x2, y1:y2].copy()
            self.frangi = image_obj.mask_frangi[x1:x2, y1:y2].copy()
            self.mask_clumpy_zone = image_obj.mask_clumpy_zone[x1:x2, y1:y2].copy()
            self.pixel_microns = image_obj.pixel_microns
            self.DoG = image_obj.mask_DoG[x1:x2, y1:y2].copy()
            self.RoG = image_obj.mask_RoG[x1:x2, y1:y2].copy()
            self.config = image_obj.config['cluster']
            """
            unused stats, might be integrated in the future
            """

            self.global_mask_background_median = image_obj.image_background_median[image_obj.mask_channel_name]
            self.global_mask_foreground_median = image_obj.image_foreground_median[image_obj.mask_channel_name]
            self.global_mask_background_std = image_obj.image_background_std[image_obj.mask_channel_name]
            self.global_mask_foreground_std = image_obj.image_foreground_std[image_obj.mask_channel_name]

            self.touching_edge = touching_edge(image_obj.shape, self.bbox)

    def create_seeds(self,
                     image_obj,
                     apply_median_filter=False):

        """
        This function infers core body parts of cells from the shape-index surface curvature profiles.
        The inferred body parts are used as seeds for watershed segmentation, the segmented particles are subsequently
        subdivided or merged by a collection of criterion.

        :param image_obj: master Image object
        :param apply_median_filter: suppresses local curvature extremities using a median kernel if set to True

        """

        """
        Import configurations
        """
        shape_index_low_bound = int(self.config['shape_index_low_bound'])
        shape_index_high_bound = int(self.config['shape_index_high_bound'])
        frangi_low_bound = float(self.config['frangi_low_bound'])
        frangi_high_bound = float(self.config['frangi_high_bound'])
        min_seed_size = int(self.config['seed_size_min'])
        radius = int(self.config['delink_radius'])
        use_frangi = bool(int(self.config['use_frangi']))

        """
        Inherit the binary mask from the master Image object
        """
        x1, y1, x2, y2 = self.bbox
        self.mask = (image_obj.mask_labeled_clusters[x1:x2, y1:y2] == self.label) * 1

        if apply_median_filter:
            # use smoothed shape-indexed image
            self.shape_indexed = image_obj.mask_shape_indexed_smoothed[x1:x2, y1:y2].copy() * self.mask
        else:
            self.shape_indexed = image_obj.mask_shape_indexed[x1:x2, y1:y2].copy() * self.mask

        """
        generate watershed seeds, remove the ones that are too small
        """
        if use_frangi:
            frangi = self.frangi*self.mask
            init_seeds = (frangi < frangi_high_bound) & (frangi > frangi_low_bound)
        else:
            init_seeds = (self.shape_indexed < shape_index_high_bound) & \
                         (self.shape_indexed > shape_index_low_bound)
        init_seeds = morphology.remove_small_holes(init_seeds, area_threshold=30) * 1
        init_seeds = morphology.opening(init_seeds, morphology.disk(radius)).astype(bool)
        init_seeds = morphology.remove_small_objects(init_seeds, min_size=min_seed_size)

        # label particles
        self.labeled_seeds = sk.measure.label(init_seeds, connectivity=1)

    def create_seeds_from_branched_particle(self, cell):
        """
        This function forces the segmentation of particles of complex shapes (eg. branched) using more stringent
        segmentation criterion.

        Note that the current solution performs suboptimally and often yields inaccurate segmentation
        Future adjustments with supervised SVM or NN are needed to reach higher segmentation accuracy and to
        account for cells with deformed morphology.

        :param cell: Cell object to be subdivided (re-imported as a Cluster object)
        :return:
        """
        self.shape_indexed = cell.shape_index
        self.label = cell.cluster_label
        self.bbox = cell.bbox
        self.phase = cell.data[cell.mask_channel_name]
        self.sobel = cell.sobel
        self.mask_clumpy_zone = np.zeros(cell.shape_index.shape)
        self.pixel_microns = cell.pixel_microns
        self.config = cell.config['cluster']
        self.mask = cell.mask * 1
        self.frangi = cell.frangi

        """
        import configurations for Cel subdivision.
        """
        use_frangi = bool(int(self.config['use_frangi_split_branch']))
        radius = int(self.config['delink_radius'])

        if not use_frangi:
            ref_image = self.shape_indexed
            low_th = int(self.config['shape_index_branch_threshold_low_bound'])
            high_th = int(self.config['shape_index_branch_threshold_high_bound'])
        else:
            ref_image = self.frangi
            low_th = float(self.config['frangi_branched_low_bound'])
            high_th = float(self.config['frangi_branched_high_bound'])
        min_seed_size = int(self.config['seed_size_min'])

        init_seeds = (ref_image < high_th) & (ref_image > low_th)
        init_seeds = morphology.remove_small_holes(init_seeds, area_threshold=30) * 1
        init_seeds = morphology.opening(init_seeds, morphology.disk(radius)).astype(bool)
        init_seeds = morphology.remove_small_objects(init_seeds, min_size=min_seed_size)
        self.labeled_seeds = measure.label(init_seeds, connectivity=1)
        del init_seeds

    def create_seeds_from_vsnap(self, cell):
        """
        This function forces the segmentation of V-shaped particles

        Note that the current solution performs suboptimally and often yields inaccurate segmentation
        Future adjustments with supervised SVM or NN are needed to reach higher segmentation accuracy and to
        account for cells with deformed morphology.

        :param cell: Cell object to be subdivided (re-imported as a Cluster object)
        :return:
        """
        self.shape_indexed = cell.shape_index
        self.label = cell.cluster_label
        self.bbox = cell.bbox
        self.phase = cell.data[cell.mask_channel_name]
        self.sobel = cell.sobel
        self.mask_clumpy_zone = np.zeros(cell.shape_index.shape)
        self.pixel_microns = cell.pixel_microns
        self.config = cell.config['cluster']
        self.mask = cell.mask * 1
        self.frangi = cell.frangi

        """
        import configurations for Cel subdivision.
        """
        use_frangi = bool(int(self.config['use_frangi_split_branch']))
        radius = int(self.config['delink_radius'])

        if not use_frangi:
            ref_image = self.shape_indexed
            low_th = int(self.config['shape_index_vshape_low_bound'])
            high_th = int(self.config['shape_index_vshape_high_bound'])
        else:
            ref_image = self.frangi
            low_th = float(self.config['frangi_branched_low_bound'])
            high_th = float(self.config['frangi_branched_high_bound'])
        min_seed_size = int(self.config['seed_size_min'])

        init_seeds = (ref_image < high_th) & (ref_image > low_th)
        init_seeds = morphology.remove_small_holes(init_seeds, area_threshold=30) * 1
        init_seeds = morphology.opening(init_seeds, morphology.disk(radius)).astype(bool)
        init_seeds = morphology.remove_small_objects(init_seeds, min_size=min_seed_size)
        self.labeled_seeds = measure.label(init_seeds, connectivity=1)
        del init_seeds

    def segmentation(self):
        """
        Cluster segmentation function.

        Given a set of inferred seed(s) (self.labeled_seeds), this function uses the watershed algorithm to
        conduct an exhaustive segmentation on the binary mask of a Cluster. The segmented Cluster may yield one
        (directly inherited by a Cell object) or multiple particles. In the latter case, Boundary objects are
        extracted from the labeled segments and are vetted for their authenticity. False boundaries are removed and
        their corresponding segments are conflated.

        """
        self.particle_counts = self.labeled_seeds.max() # 10+ times faster than computing np.unique()

        if self.particle_counts >= 1:

            if self.particle_counts == 1:
                # single segment
                if not self.touching_edge:
                    # remove particles that are too close to the edge
                    self.watershed = self.mask
                    self.particles[1] = measure.regionprops(self.mask)[0]
                else:
                    self.discarded = True

            else:
                # multple segments
                # discard segments that are close to a "clumpy" zone (by DoG)
                _clump_vs_cluster = self.mask_clumpy_zone[np.where(self.mask>0)]
                if len(np.nonzero(_clump_vs_cluster)[0])/len(_clump_vs_cluster) > 0.9:
                    self.discarded = True
                else:
                    # watershed segmentation
                    self.watershed = segmentation.watershed(self.phase, self.labeled_seeds,
                                                            mask=self.mask,
                                                            connectivity=1,
                                                            compactness=0,
                                                            watershed_line=True)
                    particle_info = measure.regionprops(self.watershed, intensity_image=self.phase)
                    for part in particle_info:
                        self.particles[part.label] = part

                    # record cell boundaries
                    self.boundary_mask = ((self.mask > 0) * 1) - ((self.watershed > 0) * 1)
                    d1, d2 = boundary.boundary_neighbor_pairwise(self.watershed, self.boundary_mask)
                    self.particle_pairwise_dict = d1
                    self.boundary_pairwise_dict = d2
                    del d1, d2

        else:
            self.discarded = True

    def compute_boundary_metrics(self, image_obj):
        """
        Analyze extracted boundaries.
        """
        for label, boundary in self.boundary_pairwise_dict.items():
            boundary.primary_boundary_metrics(self)
            image_obj.boundary_stats.append(list(boundary.metrics.values()))

    def boundary_classification(self,
                                counter,
                                boundary_prediction):
        """
        Remove low-confidence cell-cell boundaries. Merge neighboring segments/particles
        :param method:
        :return:
        """

        self.false_boundary_neighbors = []
        self.false_boundary_id = []
        for boundary_id, boundary in self.boundary_pairwise_dict.items():
            if boundary_prediction[counter]==1:
                p1, p2 = boundary.p1, boundary.p2
                self.false_boundary_neighbors.append([p1, p2])
                self.false_boundary_id.append(boundary_id)
            counter += 1
        return counter

    def remove_false_boundaries(self):
        self.merge_particle_info = merge_joint(self.false_boundary_neighbors,
                                               self.false_boundary_id)
        particle_list, boundary_list = self.merge_particle_info.newdata,\
                                       self.merge_particle_info.new_idxlist

        if len(particle_list) != len(boundary_list):
            raise ValueError('Boundary ID allocation failed!')
        else:
            for i in range(len(particle_list)):
                group = sorted(particle_list[i])
                newlabel = int(group[0])
                for label in group[1:]:
                    self.watershed[self.watershed==label] = newlabel
                    del self.particles[label]
                for boundary_id in boundary_list[i]:
                    xcoords, ycoords = np.array(self.boundary_pairwise_dict[boundary_id].boundary_coordinates).T
                    self.watershed[xcoords, ycoords] = newlabel
                self.particles[newlabel] = measure.regionprops((self.watershed == newlabel).astype(np.int),
                                                               intensity_image=self.phase)[0]

    def filter_particles(self, image_obj):
        """
        primary filtering function that removes particles that are too small and/or found in the "clumpy" zone.
        :param image_obj: master Image object
        """

        """
        minimal particle size filter
        """
        min_area = int(self.config['area_pixel_low'])
        max_area = int(self.config['area_pixel_high'])
        max_width = float(self.config['width_high'])/self.pixel_microns

        for label, particle in self.particles.items():
            _clump_area = self.mask_clumpy_zone[np.where(self.watershed == label)]

            # some basic discriminators
            in_clump = (len(np.nonzero(_clump_area))/len(_clump_area) > 0.01)*1
            too_small = (particle.area < min_area)*1
            too_large = (particle.area > max_area)*1
            too_bulky = ((particle.solidity > 0.75) & \
                         (particle.minor_axis_length > 3*max_width))*1
            # modification needed
            if np.sum([in_clump, too_small,
                       too_large, too_bulky]) == 0:
                x0, y0, _x0, _y0 = self.bbox
                x3, y3, x4, y4 = optimize_bbox(self.phase.shape, particle.bbox)
                optimized_bbox = [x3 + x0, y3 + y0, x4 + x0, y4 + y0]
                if not touching_edge(image_obj.shape, optimized_bbox):
                    optimized_mask = self.watershed[x3:x4, y3:y4] == label
                    # make a cell object
                    cell = Cell(image_obj, self.label, str(label),
                                optimized_mask, optimized_bbox, regionprop=particle)
                    cell.find_contour()
                    cell.extract_skeleton()
                    self.cells[label] = cell

    def split_branches(self, image_obj):

        """
        This function forces the segmentation of Cell objects of complex shapes
        and iteratively updates the Cluster.cells dictionary.
        :param image_obj:
        :return:
        """

        # temporary label list and cells dictionary
        _old_labels_2b_removed = []
        sub_dict = {}
        min_area = int(self.config['area_pixel_low'])
        if not self.cells == {}:
            max_label = max(list(self.cells.keys()))+1
            for label, cell in self.cells.items():

                # forced segmentation of branched object
                if cell.branched and not cell.discarded:

                    # directly transfer data from the branched Cell object to make a Cluster
                    mini_cluster = Cluster(image_obj=image_obj,
                                           cluster_regionprop=cell.regionprop)

                    # subdivide cell

                    if cell.vshaped:
                        mini_cluster.create_seeds_from_vsnap(cell)
                    else:
                        mini_cluster.create_seeds_from_branched_particle(cell)

                    # Standard cluster segmentation and boundary optimization
                    mini_cluster.segmentation()
                    if len(self.boundary_pairwise_dict)> 0:
                        for boundary_id, boundary in self.boundary_pairwise_dict.items():
                            if len(boundary.metrics.values()) > 0:
                                metrics = np.array(list(boundary.metrics.values())).reshape(1, -1)
                                normalized_metrics = image_obj.boundary_normalizer.transform(metrics)
                                false_boundary = image_obj.boundary_prediction_model.predict(normalized_metrics)[0]
                                boundary.false_boundary = bool(false_boundary)
                    mini_cluster.remove_false_boundaries()

                    # append updated Cell objects to the Cluster.cells dictionary
                    for minilabel, miniparticle in mini_cluster.particles.items():
                        x0, y0, _x0, _y0 = mini_cluster.bbox
                        x3, y3, x4, y4 = optimize_bbox(mini_cluster.phase.shape, miniparticle.bbox)
                        optimized_bbox = [x3 + x0, y3 + y0, x4 + x0, y4 + y0]
                        optimized_mask = mini_cluster.watershed[x3:x4, y3:y4] == minilabel
                        if miniparticle.area >= min_area:
                            newcell = Cell(image_obj, self.label, max_label,
                                           optimized_mask,
                                           optimized_bbox,
                                           regionprop=miniparticle)
                            newcell.find_contour()
                            newcell.extract_skeleton()
                            sub_dict[max_label] = newcell
                            max_label += 1
                            if label not in _old_labels_2b_removed:
                                _old_labels_2b_removed.append(label)
            for key in _old_labels_2b_removed:
                del self.cells[key]
            self.cells.update(sub_dict)

    def measure_cells(self,
                      polar_shapeindex=False,
                      reject_low_quality_cells=False):
        """
        Multi-metrics quantification of cellular data
        discard low-confidence Cell objects by polar shapeindex measure when specified

        :param polar_shapeindex: record low-confidence Cell objects when True
        :param reject_low_quality_cells: reject low-confidence Cell objects when True
        :return:
        """

        threshold = int(self.config['polar_shape_index_threshold'])
        for label, cell in self.cells.items():
            if not cell.branched and not cell.discarded:
                # Multi-metrics quantification
                cell.compiled_cell_process()

                # note that the Cell.compiled_cell_process may alter the Cell.branched
                # or the Cell.discarded status
                if not cell.branched and not cell.discarded:
                    if polar_shapeindex:
                        # calculates polar shape_index scores
                        # Cells with both poles correctly located should have a pole-proximate
                        # shapeindex lower than the specified threshold (strongly curved)
                        axial_shapeindex = cell.measurements.signal_measurements['Shape_index']['axial']
                        axial_shapeindex = np.interp(np.linspace(0, 1, 100),
                                                     np.linspace(0, 1, len(axial_shapeindex)),
                                                     axial_shapeindex)
                        below_th = np.where(axial_shapeindex < threshold)[0]
                        if len(below_th) >= 2:
                            if below_th[0] > 10 and below_th[-1] < 90:
                                cell.segmentation_quality_by_shapeindex = 0
                                if reject_low_quality_cells:
                                    cell.discarded = True
                        else:
                            cell.segmentation_quality_by_shapeindex = 0
                            if reject_low_quality_cells:
                                cell.discarded = True


import OMEGA_core as om
import numpy as np
import pandas as pd
from skimage.transform import rescale, resize

from skimage.transform import rescale, resize

def pole_aware_resize(data,
                      length,
                      reorientation=True,
                      nrows=20,
                      ncols=50,
                      npole=4,
                      pole_length=0.3,# unit: micron
                      flatten=True,
                      smoothen=False):

    """
    This function assumes length-independence of the polar hemisphere associated signals.
    - - + + + + + + + + - -
    - - + + + + + + + + - -
    - - + + + + + + + + - -
    - - + + + + + + + + - -
    - - + + + + + + + + - -
    - - + + + + + + + + - -



    :param data:
    :param length:
    :param reorientation:
    :param nrows:
    :param ncols:
    :param npole:
    :param flatten:
    :param smoothen:
    :return:
    """
    from skimage import filters
    if reorientation:
        half_l = int(data.shape[1] / 2)
        if np.average(data[:, :half_l]) < np.average(data[:, half_l:]):
            data = np.flip(data, axis=1)
    upscaled = rescale(data, (2, 2), anti_aliasing=True)
    pole_length = int(round((pole_length / length) * upscaled.shape[1]))
    pole1 = resize(upscaled[:, :pole_length], (nrows, npole), anti_aliasing=True)
    cell_body = resize(upscaled[:, pole_length:-pole_length], (nrows, ncols - 2 * npole), anti_aliasing=True)
    pole2 = resize(upscaled[:, -pole_length:], (nrows, npole), anti_aliasing=True)
    stitched = np.concatenate([pole1, cell_body, pole2], axis=1)
    if smoothen:
        stitched = filters.gaussian(stitched, sigma=1)
    stitched = stitched / stitched.mean()
    stitched[stitched < 0] = 0

    if flatten:
        return stitched.flatten(), length
    else:
        return stitched, length


def length_based_stack(cellpics, lengths, n=13, pad=0, remove_extreme_perc=0.05):
    extrema = max(5, int(len(lengths) * remove_extreme_perc))
    sorted_stack = cellpics.copy()[np.argsort(lengths)][extrema:-extrema]
    lengths_sorted = np.sort(lengths)[extrema:-extrema]
    stack_average = []
    lengths_average = []

    for i in range(n):
        id1 = int(i * (len(lengths_sorted) / n))
        id2 = int((i + 1) * (len(lengths_sorted) / n)) + pad
        averaged = np.average(sorted_stack[id1:id2], axis=0)
        stack_average.append(averaged)
        lengths_average.append(np.average(lengths_sorted[id1:id2]))
    return np.array(stack_average), np.array(lengths_average)


def fractional_sampling(cellpics, lengths,
                        n_groups=10,
                        group_size=10,
                        remove_extreme_perc=0.05):
    extrema = max(5, int(len(lengths) * remove_extreme_perc))
    sorted_stack = cellpics.copy()[np.argsort(lengths)][extrema:-extrema]
    sorted_lengths = np.sort(lengths)[extrema:-extrema]
    stack_average = []
    lengths_average = []

    sample_size = len(sorted_stack)
    window_size = int((sample_size - group_size) / n_groups)
    for i in range(n_groups):
        id1 = window_size * i
        id2 = window_size * i + group_size
        local_averaged = np.average(sorted_stack[id1:id2], axis=0)
        local_averaged_normalized = local_averaged / local_averaged.mean()
        stack_average.append(local_averaged_normalized)
        lengths_average.append(sorted_lengths[id1:id2].mean())
    return np.array(stack_average), np.array(lengths_average)


def agglomerative_autothreshold(strain_data):
    from sklearn.cluster import AgglomerativeClustering
    from scipy.cluster.hierarchy import linkage
    from sklearn.decomposition import PCA

    # define distance threshold
    Z = linkage(strain_data, 'ward')
    threshold = max(Z[:, 2].max() * 0.9, 3)

    # agglomerative clustering
    agg_clusterer = AgglomerativeClustering(n_clusters=None,
                                            compute_full_tree=True,
                                            distance_threshold=threshold).fit(strain_data)
    # pca to reduce the N of dimensions
    pca = PCA(n_components=2)
    pca_transformed = pca.fit_transform(strain_data)

    return np.hstack([agg_clusterer.labels_.reshape(-1, 1), pca_transformed]), threshold


def moving_window_average(data, window_size=5,interpolate = False):
    if len(data) <= window_size+2:
        raise ValueError("Input array does not have enough values")
    else:
        cum_sum = np.cumsum(np.insert(data, 0, 0))
        mw_average = (cum_sum[window_size:]-cum_sum[:-window_size])/float(window_size)
        if interpolate:
            return np.interp(np.linspace(0, 1, len(data)),
                             xp=np.linspace(0, 1, len(mw_average)),
                             fp=mw_average)
        else:
            return mw_average


def GMM(data, max_ncomponent=5):
    from sklearn import mixture
    if len(data.shape) == 1:
        data = data.reshape(-1, 1)

    best_gmm = None
    lowest_bic = np.inf
    bic = []
    for n_components in range(1, max_ncomponent + 1):
        # Fit a Gaussian mixture with EM
        gmm = mixture.GaussianMixture(n_components=n_components,
                                      covariance_type='full')
        gmm.fit(data)
        bic.append(gmm.bic(data))
        if bic[-1] < lowest_bic:
            lowest_bic = bic[-1]
            best_gmm = gmm
    return best_gmm, best_gmm.fit_predict(data), bic


def GMM_prediction_reformat(data, gmm_prediction):
    if len(data.shape) == 1:
        data = data.reshape(-1, 1)
    group_mean = []
    group_idx = []
    reformatted_prediction = np.zeros(gmm_prediction.shape)
    if gmm_prediction.max() != 0:
        for i in np.unique(gmm_prediction):
            idx = np.where(gmm_prediction == i)
            mean = data[idx].mean()
            group_mean.append(mean)
            group_idx.append(idx)
    counter = 0
    for j in np.argsort(np.array(group_mean)):
        reformatted_prediction[group_idx[j]] = counter
        counter += 1
    return reformatted_prediction


__author__ = "jz-rolling"

import numpy as np
import tifffile
import nd2reader as nd2
from . import helper
from .cluster import Cluster
from skimage import filters, morphology, registration
import pickle as pk

Version = '0.1.1'


class Image:
    """
    Read image data in the format of .nd2 hyperstack or .tif hyperstack (eg. by imageJ)

    """
    def __init__(self):

        """
        default perimeters for our microscope, changed accordingly when metadata is available

        pixel_microns
        ------- distance [um] per unit pixel

        shape
        ------- image shape

        filename 
        ------- if not specified, default filename is set as the file header of the input image file

        data
        ------- dictionary that stores different 2D imaging data by there corresponding channel names

        channels
        ------- image channel names

        mask_channel_name
        ------- channel name of the mask/bright field imaging data

        mask_binary
        ------- binary mask created by adaptive thresholding (local + global)

        mask_sobel
        ------- edge energy map using Sobel filter

        mask_clumpy_zone
        ------- energy map highlighting regions where cells aggregate notoriously

        mask_shape_indexed
        ------- 8-bit shape index map. Koenderink & Doorn, 1992

        mask_shape_indexed_smoothed
        ------- 8-bit shape index map smoothed using a median filter

        mask_labeled_clusters
        ------- labeled mini-clusters

        cluster_regionprops
        ------- regionprop dict of labeled mini-clusters

        clusters
        ------- filtered clusters that contain >= 1 cell like object(s)
        """
        self.pixel_microns = 0.065
        self.shape = [2048, 2048]

        self.filename = None
        self.data = {}
        self.channels = []
        self.mask_channel_name = '' # String key to retrieve mask/bright field imaging data
        self.mask_binary = None # Binary mask created using combinatory thresholding (local + global)
        self.mask_sobel = None # Edge energy map using Sobel filter
        self.mask_frangi = None # ridge detector
        self.mask_DoG = None # DoG filter
        self.mask_RoG = None # RoG filter
        self.mask_clumpy_zone = None
        self.mask_shape_indexed = None
        self.mask_shape_indexed_smoothed = None
        self.mask_labeled_clusters = None
        self.cluster_regionprops = {}
        self.clusters = {}
        self.boundary_stats = []

        self.image_background_median = {}
        self.image_foreground_median = {}
        self.image_background_std = {}
        self.image_foreground_std = {}
        self.config = None
        self.boundary_prediction_model = None
        self.boundary_normalizer = None
        self.cell_list = []


    def read_tiff_file(self, tiff, channels=None, filename=None, mask_channel_id=-1):

        """
        import image data from .tif file format, ideally preprocessed by imageJ
        :param tiff: input .tif hyperstack file
        :param channels: user specified channel names
        :param filename: user specified file header
        :param mask_channel_id: user specified key for retrieving bright field image

        """

        # load .tiff metafile if available
        tiff_data = tifffile.TiffFile(tiff)
        tiff_info = tiff_data.imagej_metadata["Info"]
        tiff_info = helper.ImageJinfo2dict(tiff_info)
        self.pixel_microns = float(tiff_info["dCalibration"])

        if 'channels' in tiff_data.imagej_metadata:
            _nChannels = int(tiff_data.imagej_metadata["channels"])
        else:
            _nChannels = 1

        if "Ranges" in tiff_data.imagej_metadata:
            tiff_bitdepth = tiff_data.imagej_metadata["Ranges"]
        else:
            bitdepth = 2**(int(tiff_info['BitsPerPixel']))-1
            tiff_bitdepth = []
            for i in range(_nChannels):
                tiff_bitdepth.append(0)
                tiff_bitdepth.append(bitdepth)

        if filename == None:
            self.filename = tiff.split("/")[-1]
        else:
            self.filename = filename

        if channels == None:
            channels = ['C{}'.format(x) for x in range(1,_nChannels+1)]
            self.channels = channels

        elif len(channels) != _nChannels:
            raise ValueError("The number of channels should match that of the input hyperstack.")

        self.mask_channel_name = self.channels[mask_channel_id]

        for i in range(_nChannels):
            fl_bit = tiff_bitdepth[2 * i + 1]
            self.data[channels[i]] = ((tiff_data.asarray(i) / fl_bit) * 65535).astype(np.uint16)

        # load image shape
        self.shape = self.data[self.mask_channel_name].shape

    def read_nd2_file(self, nd2file, filename=None, mask_channel_id=-1):

        """
        import image data directly from .nd2 file format
        :param nd2file: input .nd2 hyperstack image
        :param filename: user specified file header
        :param mask_channel_id: user specified key for retrieving bright field image

        """

        if filename == None:
            self.filename = nd2file.split('/')[-1]
        else:
            self.filename = filename

        # import data directly from .nd2 files
        if not nd2file.endswith(".nd2"):
            raise ValueError("Illegal input: {} is not a .nd2 file, try .read_tiff_file instead.".format(nd2file))
        nd2data = nd2.ND2Reader(nd2file)
        self.pixel_microns = nd2data.metadata["pixel_microns"]
        self.channels = nd2data.metadata["channels"]
        self.filename = filename
        self.mask_channel_name = self.channels[mask_channel_id]

        for channel in self.channels:
            self.data[channel] = np.asarray(nd2data[self.channels.index(channel)]).astype(np.uint16)

        # load image shape
        self.shape = self.data[self.mask_channel_name].shape

    def crop_edge(self, offset_correction=False, max_drift=2.5):

        """
        crop edges off the image
        :param offset_correction: correct for xy drift between phase contrast and fluorescent images if True
        :param max_drift.
        :return:
        """

        cropped = float(self.config['image']['crop_edge'])
        if self.data is None:
            raise ValueError("No images found!")

        if 0 <= cropped < 0.4:
            crop_width = int(cropped * self.shape[1])
            crop_height = int(cropped * self.shape[0])
            w1, w2 = crop_width, self.shape[1] - crop_width
            h1, h2 = crop_height, self.shape[0] - crop_height
            self.shape = (h2 - h1, w2 - w1)
        else:
            raise ValueError('Edge fraction should be no higher than 0.4 (40% from each side)!')

        if offset_correction:
            reference_image = self.data[self.mask_channel_name]
            reference_image = 100 + reference_image.max() - reference_image
            for channel, data in self.data.items():
                if channel != self.mask_channel_name:
                    shift, error, _diff = registration.phase_cross_correlation(reference_image, data,
                                                                               upsample_factor=10)
                    if max(np.abs(shift)) <= max_drift:
                        offset_image = helper.shift_image(data, shift)
                        self.data[channel] = offset_image[h1:h2, w1:w2]
                    else:
                        self.data[channel] = self.data[channel][h1:h2, w1:w2]
                else:
                    self.data[channel] = self.data[channel][h1:h2, w1:w2]

        else:
            for channel, data in self.data.items():
                self.data[channel] = self.data[channel][h1:h2, w1:w2]

    def enhance_brightfield(self, normalize=True, gamma=1.0):

        """
        attenuates signal aberrations in phase contrast image using FFT bandpass filters
        :param normalize: adjust exposure of brightfield image
        :param gamma: user specified gamma correction value, default = 1

        """
        maskimg = self.data[self.mask_channel_name].copy()
        mask_fft = helper.fft(maskimg, subtract_mean=True)
        fft_filters = helper.bandpass_filter(pixel_microns=self.pixel_microns,
                                             img_width=self.shape[1], img_height=self.shape[0],
                                             high_pass_width=float(self.config['image']['bandpass_high']),
                                             low_pass_width=float(self.config['image']['bandpass_low']))
        fft_reconstructed = helper.fft_reconstruction(mask_fft, fft_filters)

        if normalize:
            fft_reconstructed = helper.adjust_image(fft_reconstructed, adjust_gamma=True, gamma=gamma)

        self.data[self.mask_channel_name] = fft_reconstructed

        # Sobel edge operator
        sobel_filtered = filters.gaussian(filters.sobel(fft_reconstructed), sigma=1)
        self.mask_sobel = (sobel_filtered - sobel_filtered.min() + 1) / (sobel_filtered.max() + 1)

        # Frangi ridge operator
        self.mask_frangi = filters.gaussian(filters.frangi(fft_reconstructed,
                                                           black_ridges=True,
                                                           sigmas=(0.5, 2, 0.1)), sigma=1)
        _high_sigma_smoothed = filters.gaussian(fft_reconstructed, sigma=10)
        self.mask_RoG = fft_reconstructed/_high_sigma_smoothed
        self.mask_DoG = fft_reconstructed-_high_sigma_smoothed

        del mask_fft, fft_filters, fft_reconstructed, sobel_filtered

    def enhance_fluorescence(self,
                             normalize=False,
                             adjust_gamma=False,
                             gamma=1.0):
        """
        remove fluorescence background using the rolling ball method
        :param normalize: adjust exposure and data depth
        :param adjust_gamma: apply user specified gamma correction if True
        :param gamma: user specified gama correction value, default = 1
        :return:
        """

        for channel,data in self.data.items():
            if channel != self.mask_channel_name:
                bg_subtracted = helper.rolling_ball_bg_subtraction(data)
                if normalize:
                    bg_subtracted = helper.adjust_image(bg_subtracted,adjust_gamma=adjust_gamma,gamma=gamma)
                self.data[channel] = (filters.gaussian(bg_subtracted,sigma=0.5)*65535).astype(np.uint16)
                del bg_subtracted


    def locate_clusters(self):
        """
        primary segmentation, splits binary mask into patches (Cluster)
        """

        """
        initial segmentation
        mask_binary denotes the binary mask of cells;
        mask_shape_indexed denotes the shape_indexed surface curvature map of the phase contrast image;
        mask_labeled_clusters is a labeled image of which each unique integer (1->N) denotes a patch of pixels 
        that makes a Cluster.
        """

        """
        import configurations
        """
        clump_threshold = float(self.config['image']['clump_threshold'])
        clump_sigma = float(self.config['image']['clump_sigma'])
        shape_index_sigma = float(self.config['image']['shape_index_sigma'])
        area_pixel_low = float(self.config['image']['area_pixel_low'])

        """
        primary segmentation
        """
        self.mask_binary, \
        self.mask_shape_indexed, \
        self.mask_labeled_clusters, \
        regionprops = helper.init_segmentation(self.data[self.mask_channel_name],
                                               shape_indexing_sigma=shape_index_sigma,
                                               min_particle_size=area_pixel_low)

        """
        mask_clumpy_zone takes a DoG transformed phase contrast image and marks the region where 
        cells are likely forming clumps or overlapping. 
        """
        _flattened = self.mask_RoG[np.where(self.mask_binary>0)]
        self.mask_clumpy_zone = self.mask_RoG > \
                                (np.mean(_flattened)+clump_threshold*np.std(_flattened))
        self.mask_clumpy_zone = filters.gaussian(self.mask_clumpy_zone,
                                                 sigma=clump_sigma)

        """
        create a dictionary of regional information of the binary mask
        """
        for regionprop in regionprops:
            self.cluster_regionprops[regionprop.label] = regionprop
        self.mask_shape_indexed_smoothed = filters.median(self.mask_shape_indexed,
                                                          morphology.disk(2)).astype(np.uint8)

        """
        calculate global signal metrics
        """
        for channel, data in self.data.items():
            foreground = np.where(self.mask_binary > 0)
            background = np.where(self.mask_binary == 0)
            bg_data = data[background]
            fg_data = data[foreground]
            self.image_background_median[channel] = np.median(bg_data)
            self.image_foreground_median[channel] = np.median(fg_data)
            self.image_background_std[channel] = np.std(bg_data)
            self.image_foreground_std[channel] = np.std(fg_data)
        del bg_data, fg_data, foreground, background, regionprops

    def cluster_segmentation(self, predictor, normalizer,
                             boundary_annotation=False,
                             apply_median_filter=False):
        """

        :param predictor:
        :param normalizer:
        :param boundary_annotation:
        :param apply_median_filter:
        :return:
        """

        for idx, regionprop in self.cluster_regionprops.items():
            cluster = Cluster(self, regionprop)
            cluster.create_seeds(self, apply_median_filter=apply_median_filter)
            cluster.segmentation()
            cluster.compute_boundary_metrics(self)
            self.clusters[idx] = cluster

        # boundary prediction with trained model
        if not boundary_annotation:
            self.boundary_prediction_model = pk.load(open(predictor, 'rb'))
            self.boundary_normalizer = pk.load(open(normalizer, 'rb'))
            if len(self.boundary_stats) > 0:
                if len(self.boundary_stats) == 1:
                    self.boundary_stats = np.array(self.boundary_stats).reshape(1, -1)
                norm_boundary_stats = self.boundary_normalizer.transform(self.boundary_stats)
                boundary_prediction = self.boundary_prediction_model.predict(norm_boundary_stats)
                counter = 0
                for idx, cluster in self.clusters.items():
                    counter = cluster.boundary_classification(counter, boundary_prediction)
                    cluster.remove_false_boundaries()

    def cell_segmentation(self,
                          split_branches=True,
                          shapeindex_quality=True,
                          filter_by_shapeindex_quality=True):
        """
        evaluates the accuracy of Cluster segmentation, accepts or rejects Cell objects based
        on the computed multimetrics.
        :param split_branches: forces the subdivision of a non-rod-shaped particle
        :param shapeindex_quality: additional segmentation quality assessment using polar shape index measurements.
        :param filter_by_shapeindex_quality:
        :return:
        """
        for idx, cluster in self.clusters.items():
            if not cluster.discarded:
                cluster.filter_particles(self)
                if split_branches:
                    cluster.split_branches(self)
                cluster.measure_cells(polar_shapeindex=shapeindex_quality,
                                      reject_low_quality_cells=filter_by_shapeindex_quality)

    def get_cells(self,
                  discarded=False,
                  branched=False,
                  return_list=True):
        self.cell_list=[]
        for cluster_key, cluster in self.clusters.items():
            for cell_key, cell in cluster.cells.items():
                if cell.discarded:
                    if discarded:
                        self.cell_list.append(cell)
                elif cell.branched:
                    if branched:
                        self.cell_list.append(cell)
                else:
                    self.cell_list.append(cell)

        if return_list:
            return self.cell_list


def _test_stitched_data(tiff, shape=(5, 5), channels=None,
                        filename=None, mask_channel_id=0):
    """
    temporary handle for stitched image files.
    :param tiff: input .tif hyperstack file
    :param shape: N X M stitched images
    :param channels: user specified channel names
    :param filename: user specified file header
    :param mask_channel_id: user specified key for retrieving bright field image
    """


    """
    import data with tifffile
    """
    tiff_data = tifffile.TiffFile(tiff)
    tiff_info = tiff_data.imagej_metadata

    if filename == None:
        filename = tiff.split("/")[-1]
    else:
        filename = filename

    # load .tiff metafile if available
    pixel_microns = float(tiff_info["spacing"])
    _nChannels = int(tiff_info["channels"])

    if channels == None:
        channels = ['C{}'.format(x) for x in range(1, _nChannels + 1)]

    elif len(channels) != _nChannels:
        raise ValueError("The number of channels should match that of the input hyperstack.")

    mask_channel_name = channels[mask_channel_id]
    d_data = tiff_data.asarray()
    _c, h, w= d_data.shape

    if _c != len(channels):
        raise ValueError("The number of channels should match that of the input hyperstack.")

    if (h % shape[0] + w % shape[1]) != 0:
        raise ValueError("Error found when OMEGA attempted image splitting!")

    """
    user defined splitting parameters
    """
    unit_h = int(h/shape[0])
    unit_w = int(w/shape[0])

    image_count = 0
    pooled_images = []

    for i in range(0, h, unit_h):
        for j in range(0, w, unit_w):
            img = Image()
            img.filename = '{}_{}'.format(filename, image_count)
            img.pixel_microns = pixel_microns
            img.channels = channels
            img.mask_channel_name = mask_channel_name
            img.shape = [unit_h, unit_w]
            for c in range(_nChannels):
                channel = channels[c]
                img.data[channel] = d_data[c, i:i+unit_h, j:j+unit_w]
            pooled_images.append(img)
            image_count += 1

    return pooled_images


from .helper import *
from .metrics import *


class Measurement:

    """
    The Measurement class generates multi-metrics signal profiles for each extracted Cell object.
    It contains the following
    """
    def __init__(self):
        """
        Attributes
        ==========

        signal_measurements
        ---------- image signal profiles

        morphology_measurements
        ---------- morphological profiles

        """
        #self.measure_midline = None
        #self.straighten = None
        #self.straghten_normalize_width = None
        #self.SNR = None
        self.signal_measurements = {}
        self.morphology_measurements = {}

    def signal(self, cell, channel, profiling_mesh):
        """
        signal profiling function

        :param cell: Cell object
        :param channel: channel name
        :param profiling_mesh: subpixel profiling mesh

        """
        if channel == 'Shape_index':
            data = cell.shape_index
        else:
            data = cell.data[channel]

        midline = cell.midlines[0]
        width = cell.width_lists[0]
        mask = cell.mask
        length = cell.lengths[0]

        if channel not in self.signal_measurements:
            channel_data = {}
            channel_data['axial'] = measure_along_strip(midline, data, width=5)
            channel_data['straighten'] = straighten_cell(data, midline, width)
            straighten_normalized = bilinear_interpolate_numpy(data, profiling_mesh[0], profiling_mesh[1]).T
            channel_data['straighten_normalized'] = straighten_normalized
            bg_removed = data[np.where(mask>0)]
            channel_data['median'] = np.median(bg_removed)
            channel_data['standard_deviation'] = np.std(bg_removed)
            channel_data['mean'] = np.mean(bg_removed)
            channel_data['max'] = np.max(bg_removed)
            channel_data['min'] = np.min(bg_removed)
            midline_kurtosis, midline_skewness = kurtosis_skewness(channel_data['axial'])
            channel_data['midline_skewness'] = midline_skewness
            channel_data['midline_kurtosis'] = midline_kurtosis
            cell_kurtosis, cell_skewness = kurtosis_skewness(data[np.where(mask>0)])
            channel_data['cell_kurtosis'] = cell_kurtosis
            channel_data['cell_skewness'] = cell_skewness
            channel_data['axial_symmetry'] = measure_symmetry(channel_data['axial'],weighted=True)
            channel_data['axial_mean'] = np.average(channel_data['straighten_normalized'],axis=0)
            channel_data['axial_std'] = np.std(channel_data['straighten_normalized'],axis=0)
            channel_data['lateral_mean'] = np.average(channel_data['straighten_normalized'],axis=1)
            channel_data['lateral_std'] = np.std(channel_data['straighten_normalized'], axis=1)
            FWHM, offset_score = FWHM_background_aware(channel_data['lateral_mean'])
            channel_data['lateral_FWHM'] = FWHM
            channel_data['lateral_center_offset'] = offset_score
            channel_data['normalized_lateral_FWHM'] = channel_data['lateral_FWHM']/len(channel_data['lateral_mean'])
            channel_data['lateral_symmetry'] = measure_symmetry(channel_data['lateral_mean'], weighted=True)
            channel_data['segmented_measurements'] = divede_cell_pos(channel_data['axial_mean'], length)
            self.signal_measurements[channel] = channel_data

    def update_signal(self, data,
                      channel,
                      midline,
                      width,
                      mask,
                      length,
                      profiling_mesh):
        """
        update existing data

        :param data: image data of interest
        :param channel: channel name
        :param midline: precomputed cell midline
        :param width: precomputed cell widths profile (unit: pixel)
        :param mask: precomputed cell binary mask
        :param length: cell length (unit: micron)
        :param profiling_mesh: precomputed orthogonal profiling mesh

        """
        channel_data = {}
        channel_data['axial'] = measure_along_strip(midline, data, width=5)
        channel_data['straighten'] = straighten_cell(data, midline, width)
        channel_data['straighten_normalized'] = bilinear_interpolate_numpy(data,
                                                                           profiling_mesh[0],
                                                                           profiling_mesh[1]).T
        bg_removed = data[np.where(mask > 0)]
        channel_data['median'] = np.median(bg_removed)
        channel_data['standard_deviation'] = np.std(bg_removed)
        channel_data['mean'] = np.mean(bg_removed)
        channel_data['max'] = np.max(bg_removed)
        channel_data['min'] = np.min(bg_removed)
        midline_kurtosis, midline_skewness = kurtosis_skewness(channel_data['axial'])
        channel_data['midline_skewness'] = midline_skewness
        channel_data['midline_kurtosis'] = midline_kurtosis
        cell_kurtosis, cell_skewness = kurtosis_skewness(data[np.where(mask > 0)])
        channel_data['cell_kurtosis'] = cell_kurtosis
        channel_data['cell_skewness'] = cell_skewness
        channel_data['axial_symmetry'] = measure_symmetry(channel_data['axial'], weighted=True)
        channel_data['axial_mean'] = np.average(channel_data['straighten_normalized'], axis=0)
        channel_data['axial_std'] = np.std(channel_data['straighten_normalized'], axis=0)
        channel_data['lateral_mean'] = np.average(channel_data['straighten_normalized'], axis=1)
        channel_data['lateral_std'] = np.std(channel_data['straighten_normalized'], axis=1)
        FWHM, offset_score = FWHM_background_aware(channel_data['lateral_mean'])
        channel_data['lateral_FWHM'] = FWHM
        channel_data['lateral_center_offset'] = offset_score
        channel_data['normalized_lateral_FWHM'] = channel_data['lateral_FWHM'] / len(channel_data['lateral_mean'])
        channel_data['lateral_symmetry'] = measure_symmetry(channel_data['lateral_mean'], weighted=True)
        channel_data['segmented_measurements'] = divede_cell_pos(channel_data['axial_mean'], length)
        self.signal_measurements[channel] = channel_data

    def particle_morphology(self, cell):
        """
        morphological properties of the binary mask of any extracted particle
        :param cell: corresponding Cell object
        """

        # basic morphological features
        self.morphology_measurements['area'] = cell.regionprop.area * (cell.pixel_microns**2)
        self.morphology_measurements['eccentricity'] = cell.regionprop.eccentricity
        self.morphology_measurements['aspect_ratio'] = cell.regionprop.minor_axis_length / \
                                                       cell.regionprop.major_axis_length
        self.morphology_measurements['solidity'] = cell.regionprop.solidity
        self.morphology_measurements['circularity'] = circularity(cell.regionprop)
        self.morphology_measurements['convexity'] = convexity(cell.regionprop)

        # note that here the rough contour rather than the optimized one is used
        self.morphology_measurements['average_bending_energy'] = bending_energy(cell.init_contour)
        self.morphology_measurements['normalized_bending_energy'] = normalized_contour_complexity(cell)

        angle_along_contour = bend_angle(cell.init_contour, window=5)
        # an additional 180 degrees are added to the min/max curvatures to suppress negativity
        self.morphology_measurements['min_curvature'] = (180+angle_along_contour.min())/180
        self.morphology_measurements['max_curvature'] = (180+angle_along_contour.max())/180
        self.morphology_measurements['std_curvature'] = np.abs(angle_along_contour).std()/180
        self.morphology_measurements['mean_curvature'] = np.abs(angle_along_contour).mean()/180

        #pixelated length estimation
        rough_length, major_axis_length = cell.skeleton.sum()*np.sqrt(2), cell.regionprop.major_axis_length
        self.morphology_measurements['rough_Length'] = max(rough_length, major_axis_length)*cell.pixel_microns
        self.morphology_measurements['rough_sinuosity'] = max(rough_length, major_axis_length)/major_axis_length
        if self.morphology_measurements['rough_sinuosity'] < 1:
            self.morphology_measurements['rough_sinuosity'] = 1
        self.morphology_measurements['branch_count'] = max(0, len(cell.skeleton_coords) - 2)

    def cell_morphology(self, cell):
        """
        morphological properties estimated with subpixel contour/midlines

        :param cell: Cell object

        """
        self.morphology_measurements['length'] = np.sum(cell.lengths)
        self.morphology_measurements['sinuosity'] = sinuosity(cell.midlines[0])

        # ignore widths measures from the polar 0.3 microns
        widths = np.array(cell.width_lists[0])
        dl = int(0.3*len(widths)/self.morphology_measurements['length'])
        filtered_widths = widths[dl:-dl] * cell.pixel_microns
        self.morphology_measurements['width_median'] = np.median(filtered_widths)
        self.morphology_measurements['width_std'] = np.std(filtered_widths)
        self.morphology_measurements['width_max'] = np.max(filtered_widths)
        self.morphology_measurements['width_min'] = np.max(filtered_widths)

        # find where the min and max width lies along the cell
        min_width_pos_fraction = np.argmin(filtered_widths)/len(filtered_widths)
        if min_width_pos_fraction >= 0.5:
            min_width_pos_fraction = 1-min_width_pos_fraction
        max_width_pos_fraction = np.argmax(filtered_widths)/len(filtered_widths)
        if max_width_pos_fraction >= 0.5:
            max_width_pos_fraction = 1-max_width_pos_fraction
        self.morphology_measurements['min_width_pos_fraction'] = min_width_pos_fraction
        self.morphology_measurements['max_width_pos_fraction'] = max_width_pos_fraction
        self.morphology_measurements['width_symmetry'] = measure_symmetry(filtered_widths, weighted=True)


    def data_to_dict(self, by='signal'):

        """
        export cellular metrics to data dictionaries

        :param by: selection of exported metric(s)
                   ----'signal' export only image signal metrics
                   ----'particle_morphology' export only basic morphological properties of a binary particle
                   ----'cell_morphology' export only cell-specific morphological properties
                   ----'morphology' export all morphological properties
                   ----'all' export all data
        :return: keys_dict
                 ------keys to retrieve data
                 output_dict
                 ------dictionaries corresponding to the selected metrics
        """

        signal_keys = ['median', 'mean', 'max', 'min', 'standard_deviation',
                       'midline_skewness', 'midline_kurtosis', 'cell_kurtosis', 'cell_skewness',
                       'axial_symmetry', 'lateral_FWHM', 'lateral_center_offset', 'normalized_lateral_FWHM',
                       'lateral_symmetry']

        particle_morphological_keys = ['eccentricity', 'solidity', 'circularity',
                                       'convexity', 'average_bending_energy',
                                       'min_curvature', 'max_curvature',
                                       'mean_curvature', 'std_curvature',
                                       'rough_Length', 'rough_sinuosity', 'branch_count']

        cell_morphological_keys = ['length', 'sinuosity',
                                   'width_median', 'width_std', 'width_max', 'width_min'
                                   'width_symmetry']

        output_dict = {}
        keys_dict = {}
        if by == 'signal':
            for channel in list(self.signal_measurements.keys()):
                output_dict[channel] = list(map(self.signal_measurements[channel].get, signal_keys))
                keys_dict[channel] = signal_keys
            return keys_dict, output_dict

        elif by == 'particle_morphology':
            output_dict[by] = list(map(self.morphology_measurements.get, particle_morphological_keys))
            keys_dict[by] = particle_morphological_keys
            return keys_dict, output_dict

        elif by == 'cell_morphology':
            output_dict[by] = list(map(self.morphology_measurements.get, cell_morphological_keys))
            keys_dict[by] = cell_morphological_keys
            return keys_dict, output_dict

        elif by == 'morphology':
            morphology_keys = particle_morphological_keys + cell_morphological_keys
            output_dict[by] = list(map(self.morphology_measurements.get, morphology_keys))
            keys_dict[by] = morphology_keys
            return keys_dict, output_dict

        elif by == 'all':
            for channel in list(self.signal_measurements.keys()):
                output_dict[channel] = list(map(self.signal_measurements[channel].get, signal_keys))
                keys_dict[channel] = signal_keys
            output_dict['morphology'] = list(map(self.morphology_measurements.get,
                                             particle_morphological_keys + cell_morphological_keys))
            keys_dict['morphology'] = particle_morphological_keys + cell_morphological_keys
            return keys_dict, output_dict

        else:
            text1 = "Illegal key name, try the following: "
            text2 = "'signal', 'particle_morphology', 'cell_morphology', 'morphology', 'all'."
            raise ValueError(text1+text2)

